# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import json
from typing import Dict, Union
from VeraGridEngine.IO.file_system import get_create_veragrid_folder

from VeraGrid.Gui.Main.SubClasses.base_gui import BaseMainGui
from VeraGrid.Session.server_driver import ServerDriver
from VeraGrid.Gui.messages import warning_msg, yes_no_question


class ServerMain(BaseMainGui):
    """
    MainGUI
    """

    def __init__(self, parent=None) -> None:
        """
        Main constructor
        """

        # create main window
        BaseMainGui.__init__(self, parent=parent)

        # Server driver
        self.server_driver: ServerDriver = ServerDriver(url="", port=0, pwd="", secure=False)
        self.server_driver.connected_signal.connect(self.server_connected)
        self.server_driver.done_signal.connect(self.post_start_stop_server)  # connect the post function

        self.ui.server_tableView.setModel(self.server_driver.data_model)

        # menu
        self.ui.actionEnable_server_mode.triggered.connect(self.server_start_stop)

        # table double click
        self.ui.server_tableView.doubleClicked.connect(self.get_results)

    @staticmethod
    def server_config_file_path() -> str:
        """
        get the config file path
        :return: config file path
        """
        return os.path.join(get_create_veragrid_folder(), 'server_config.json')

    def server_config_file_exists(self) -> bool:
        """
        Check if the config file exists
        :return: True / False
        """
        return os.path.exists(self.server_config_file_path())

    def get_gui_server_config_data(self):
        """
        Get server data from the GUI
        :return:
        """
        return {"url": self.ui.server_url_lineEdit.text(),
                "port": self.ui.server_port_spinBox.value(),
                "user": "",
                "pwd": self.ui.server_pwd_lineEdit.text(),
                "secure": self.ui.secureServerConnectionCheckBox.isChecked(),}

    def save_server_config(self):
        """
        Save the GUI configuration
        :return:
        """
        data = self.get_gui_server_config_data()
        with open(self.server_config_file_path(), "w") as f:
            f.write(json.dumps(data, indent=4))

    def apply_server_config(self, data: Dict[str, Union[str, int]]) -> None:
        """
        Apply the server config
        :param data: Some local data
        """
        self.ui.server_url_lineEdit.setText(data.get("url", "localhost"))
        self.ui.server_port_spinBox.setValue(data.get("port", 8080))
        # "user": "",
        self.ui.server_pwd_lineEdit.setText(data.get("pwd", "1234"))
        self.ui.secureServerConnectionCheckBox.setChecked(data.get("secure", True))

    def load_server_config(self) -> None:
        """
        Load server configuration from the local user folder
        """
        if self.server_config_file_exists():
            with open(self.server_config_file_path(), "r") as f:
                try:
                    data = json.load(f)
                    self.apply_server_config(data=data)
                except json.decoder.JSONDecodeError as e:
                    print(e)
                    self.save_server_config()
                    self.show_error_toast("Server config file was erroneous, wrote a new one")

    def server_start_stop(self):
        """

        :return:
        """
        if self.ui.actionEnable_server_mode.isChecked():

            # create a new driver
            self.server_driver.set_values(url=self.ui.server_url_lineEdit.text().strip(),
                                          port=self.ui.server_port_spinBox.value(),
                                          pwd=self.ui.server_pwd_lineEdit.text().strip(),
                                          secure=self.ui.secureServerConnectionCheckBox.isChecked(),
                                          status_func=self.ui.server_status_label.setText)

            # save the last server config
            self.save_server_config()

            # run asynchronously
            self.server_driver.start()

        else:

            ok = yes_no_question(text="The server connection is running, are you sure that you want to stop it?",
                                 title="Stop Server")

            if ok:
                self.server_driver.cancel()
                self.ui.actionEnable_server_mode.setChecked(False)
            else:
                self.ui.actionEnable_server_mode.setChecked(True)

    def post_start_stop_server(self):
        """
        Post server run
        :return:
        """
        if not self.server_driver.is_running():
            if self.server_driver.logger.has_logs():
                self.show_error_toast(message="Could not connect to the server :/")
                self.ui.actionEnable_server_mode.setChecked(False)

    def server_connected(self):
        """
        Called upon server connection
        :return:
        """
        self.show_info_toast(message="Connected!")


    def get_results(self):
        """

        :return:
        """

        indices = self.ui.server_tableView.selectedIndexes()

        if len(indices) == 1:

            row_idx = indices[0].row()

            job = self.server_driver.data_model.jobs[row_idx]

            self.server_driver.download_results(job_id=job.id_tag,
                                                api_key="",
                                                local_filename=job.id_tag + '.results')

        self.show_info_toast("results received!")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'block_data_dialogue.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QFrame, QHBoxLayout,
    QHeaderView, QListWidget, QListWidgetItem, QSizePolicy,
    QSplitter, QTableView, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(925, 514)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.splitter_3 = QSplitter(Dialog)
        self.splitter_3.setObjectName(u"splitter_3")
        self.splitter_3.setOrientation(Qt.Orientation.Horizontal)
        self.frame_8 = QFrame(self.splitter_3)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setMaximumSize(QSize(400, 16777215))
        self.frame_8.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(-1, 8, -1, -1)
        self.datalistWidget = QListWidget(self.frame_8)
        brush = QBrush(QColor(64, 191, 83, 255))
        brush.setStyle(Qt.BrushStyle.NoBrush)
        __qlistwidgetitem = QListWidgetItem(self.datalistWidget)
        __qlistwidgetitem.setForeground(brush);
        brush1 = QBrush(QColor(64, 191, 83, 255))
        brush1.setStyle(Qt.BrushStyle.NoBrush)
        __qlistwidgetitem1 = QListWidgetItem(self.datalistWidget)
        __qlistwidgetitem1.setForeground(brush1);
        brush2 = QBrush(QColor(26, 95, 180, 255))
        brush2.setStyle(Qt.BrushStyle.NoBrush)
        __qlistwidgetitem2 = QListWidgetItem(self.datalistWidget)
        __qlistwidgetitem2.setForeground(brush2);
        brush3 = QBrush(QColor(26, 95, 180, 255))
        brush3.setStyle(Qt.BrushStyle.NoBrush)
        __qlistwidgetitem3 = QListWidgetItem(self.datalistWidget)
        __qlistwidgetitem3.setForeground(brush3);
        brush4 = QBrush(QColor(255, 120, 0, 255))
        brush4.setStyle(Qt.BrushStyle.NoBrush)
        __qlistwidgetitem4 = QListWidgetItem(self.datalistWidget)
        __qlistwidgetitem4.setForeground(brush4);
        self.datalistWidget.setObjectName(u"datalistWidget")

        self.verticalLayout_7.addWidget(self.datalistWidget)

        self.splitter_3.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.splitter_3)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.Shape.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.splitter = QSplitter(self.PlotFrame)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.tableView_variables_and_params = QTableView(self.splitter)
        self.tableView_variables_and_params.setObjectName(u"tableView_variables_and_params")
        self.splitter.addWidget(self.tableView_variables_and_params)
        self.tableView_equations = QTableView(self.splitter)
        self.tableView_equations.setObjectName(u"tableView_equations")
        self.splitter.addWidget(self.tableView_equations)

        self.horizontalLayout.addWidget(self.splitter)

        self.splitter_3.addWidget(self.PlotFrame)

        self.verticalLayout.addWidget(self.splitter_3)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))

        __sortingEnabled = self.datalistWidget.isSortingEnabled()
        self.datalistWidget.setSortingEnabled(False)
        ___qlistwidgetitem = self.datalistWidget.item(0)
        ___qlistwidgetitem.setText(QCoreApplication.translate("Dialog", u"State variables", None));
        ___qlistwidgetitem1 = self.datalistWidget.item(1)
        ___qlistwidgetitem1.setText(QCoreApplication.translate("Dialog", u"State equations", None));
        ___qlistwidgetitem2 = self.datalistWidget.item(2)
        ___qlistwidgetitem2.setText(QCoreApplication.translate("Dialog", u"Algebraic variables", None));
        ___qlistwidgetitem3 = self.datalistWidget.item(3)
        ___qlistwidgetitem3.setText(QCoreApplication.translate("Dialog", u"Algebraic equations", None));
        ___qlistwidgetitem4 = self.datalistWidget.item(4)
        ___qlistwidgetitem4.setText(QCoreApplication.translate("Dialog", u"Parameters", None));
        self.datalistWidget.setSortingEnabled(__sortingEnabled)

    # retranslateUi






# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING, Union
from PySide6.QtGui import QIcon, QPixmap
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGrid.Gui.messages import yes_no_question
from VeraGridEngine.Devices.Branches.winding import Winding
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class WindingGraphicItem(LineGraphicTemplateItem):

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem, None],
                 editor: SchematicWidget,
                 width=5,
                 api_object: Winding = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """
        LineGraphicTemplateItem.__init__(self=self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         draw_labels=draw_labels)

        self.parent_tr3_graphics_item = None
        self.winding_number = 0

    @property
    def api_object(self) -> Winding:
        return self._api_object

    @api_object.setter
    def api_object(self, api_object: Winding):
        self._api_object = api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("Winding")

            pe = menu.addAction('Active')
            pe.setCheckable(True)
            pe.setChecked(self.api_object.active)
            pe.triggered.connect(self.enable_disable_toggle)

            add_menu_entry(menu=menu,
                           text="Draw labels",
                           function_ptr=self.enable_disable_label_drawing,
                           checkeable=True,
                           checked_value=self.draw_labels)

            # menu.addSeparator()

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.png"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)

            # menu.addSeparator()

            ra2 = menu.addAction('Delete')
            del_icon = QIcon()
            del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.png"))
            ra2.setIcon(del_icon)
            ra2.triggered.connect(self.delete)

            menu.exec_(event.screenPos())
        else:
            pass

    def delete(self, ask=True):
        """
        Remove this object in the diagram and the API
        @return:
        """
        deleted, delete_from_db_final = self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)

        if deleted:
            # self.editor.circuit.delete_branch(self.api_object)
            # self.editor.delete_element_utility_function(self.api_object)

            # unregister the winding
            self.parent_tr3_graphics_item.remove_winding(self.winding_number)





# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'substation_designer_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QDoubleSpinBox, QFrame,
    QGridLayout, QHBoxLayout, QHeaderView, QLabel,
    QLineEdit, QPushButton, QSizePolicy, QSpacerItem,
    QTableView, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(707, 359)
        self.gridLayout = QGridLayout(Dialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.londoubleSpinBox = QDoubleSpinBox(Dialog)
        self.londoubleSpinBox.setObjectName(u"londoubleSpinBox")
        font = QFont()
        font.setPointSize(10)
        self.londoubleSpinBox.setFont(font)
        self.londoubleSpinBox.setDecimals(4)

        self.gridLayout.addWidget(self.londoubleSpinBox, 1, 3, 1, 1)

        self.label_2 = QLabel(Dialog)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setFont(font)

        self.gridLayout.addWidget(self.label_2, 0, 2, 1, 1)

        self.label_4 = QLabel(Dialog)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setFont(font)

        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)

        self.se_name_lineEdit = QLineEdit(Dialog)
        self.se_name_lineEdit.setObjectName(u"se_name_lineEdit")
        self.se_name_lineEdit.setFont(font)

        self.gridLayout.addWidget(self.se_name_lineEdit, 0, 1, 1, 1)

        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.gridLayout_2 = QGridLayout(self.frame)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)

        self.gridLayout.addWidget(self.frame, 4, 0, 1, 2)

        self.latdoubleSpinBox = QDoubleSpinBox(Dialog)
        self.latdoubleSpinBox.setObjectName(u"latdoubleSpinBox")
        self.latdoubleSpinBox.setFont(font)
        self.latdoubleSpinBox.setDecimals(4)

        self.gridLayout.addWidget(self.latdoubleSpinBox, 1, 1, 1, 1)

        self.se_code_lineEdit = QLineEdit(Dialog)
        self.se_code_lineEdit.setObjectName(u"se_code_lineEdit")
        self.se_code_lineEdit.setFont(font)

        self.gridLayout.addWidget(self.se_code_lineEdit, 0, 3, 1, 1)

        self.label_5 = QLabel(Dialog)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setFont(font)

        self.gridLayout.addWidget(self.label_5, 1, 2, 1, 1)

        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")
        self.label.setFont(font)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.tableView = QTableView(Dialog)
        self.tableView.setObjectName(u"tableView")
        self.tableView.setFont(font)

        self.gridLayout.addWidget(self.tableView, 3, 0, 1, 4)

        self.frame_2 = QFrame(Dialog)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.addVlButton = QPushButton(self.frame_2)
        self.addVlButton.setObjectName(u"addVlButton")
        self.addVlButton.setFont(font)

        self.horizontalLayout.addWidget(self.addVlButton)

        self.deleteVlButton = QPushButton(self.frame_2)
        self.deleteVlButton.setObjectName(u"deleteVlButton")
        self.deleteVlButton.setFont(font)

        self.horizontalLayout.addWidget(self.deleteVlButton)

        self.horizontalSpacer = QSpacerItem(171, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.createButton = QPushButton(self.frame_2)
        self.createButton.setObjectName(u"createButton")
        self.createButton.setMaximumSize(QSize(80, 16777215))
        self.createButton.setFont(font)

        self.horizontalLayout.addWidget(self.createButton)


        self.gridLayout.addWidget(self.frame_2, 5, 0, 1, 4)

        self.label_3 = QLabel(Dialog)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setFont(font)

        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 4)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
        self.londoubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" \u00ba", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Code", None))
        self.label_4.setText(QCoreApplication.translate("Dialog", u"Latitude", None))
        self.latdoubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" \u00ba", None))
        self.label_5.setText(QCoreApplication.translate("Dialog", u"Longitude", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Name", None))
#if QT_CONFIG(tooltip)
        self.addVlButton.setToolTip(QCoreApplication.translate("Dialog", u"Add voltage level", None))
#endif // QT_CONFIG(tooltip)
        self.addVlButton.setText(QCoreApplication.translate("Dialog", u"+", None))
#if QT_CONFIG(tooltip)
        self.deleteVlButton.setToolTip(QCoreApplication.translate("Dialog", u"Remove voltage level", None))
#endif // QT_CONFIG(tooltip)
        self.deleteVlButton.setText(QCoreApplication.translate("Dialog", u"-", None))
#if QT_CONFIG(tooltip)
        self.createButton.setToolTip(QCoreApplication.translate("Dialog", u"Create substation", None))
#endif // QT_CONFIG(tooltip)
        self.createButton.setText(QCoreApplication.translate("Dialog", u"Create", None))
        self.label_3.setText(QCoreApplication.translate("Dialog", u"Voltage levels", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, List, Union
from PySide6.QtCore import Qt, QPointF
from PySide6.QtGui import QPen, QCursor, QColor, QBrush
from PySide6.QtWidgets import (QGraphicsEllipseItem, QMenu,
                               QGraphicsSceneContextMenuEvent,
                               QGraphicsSceneMouseEvent)
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.MapWidget.Substation.node_template import NodeTemplate

from VeraGridEngine.Devices.types import INJECTION_DEVICE_TYPES

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget


class MapInjectionTemplateGraphicItem(NodeTemplate, QGraphicsEllipseItem):
    """
    InjectionTemplateGraphicItem
    """

    def __init__(self,
                 api_object: INJECTION_DEVICE_TYPES,
                 editor: GridMapWidget,
                 lat: float,
                 lon: float,
                 size: float = 0.8,
                 draw_labels: bool = True):
        """

        :param api_object:
        :param editor:
        :param lat:
        :param lon:
        :param size:
        :param draw_labels:
        """
        # Correct way to call multiple inheritance
        super().__init__()

        # Explicitly call QGraphicsRectItem initialization
        QGraphicsEllipseItem.__init__(self)

        # Explicitly call NodeTemplate initialization
        NodeTemplate.__init__(self,
                              api_object=api_object,
                              editor=editor,
                              draw_labels=draw_labels,
                              lat=lat,
                              lon=lon)

        r2 = size / 2.0
        x, y = editor.to_x_y(lat=lat, lon=lon)  # upper left corner
        self.size = size
        self.setRect(x - r2, y - r2, self.size, self.size)

        # Properties of the container:
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsMovable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.set_api_object_color()

    @property
    def api_object(self) -> INJECTION_DEVICE_TYPES:
        """

        :return:
        """
        return self._api_object

    @property
    def editor(self) -> GridMapWidget:
        """

        :return:
        """
        return self._editor

    def set_api_object_color(self) -> QColor:
        """
        Sets whatever colour is in the API object property
        """

        color = QColor(self.api_object.color)
        color.setAlpha(128)
        self.color = color

        brush = QBrush(color)
        self.setBrush(brush)

        pen = self.pen()
        pen.setColor(color)
        pen.setWidth(0)  # set the pen width to 0
        self.setPen(pen)

        return color

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        """
        Event handler for mouse press events.
        """
        super().mousePressEvent(event)
        # selected_items = self.editor.map.view.selected_items()
        # if len(selected_items) < 2:
        self.setSelected(True)

        event.setAccepted(True)
        self.editor.map.view.disable_move = True

        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent):
        """
        Event handler for mouse release events.
        """
        # super().mouseReleaseEvent(event)
        self.editor.disableMove = True
        self.update_position_at_the_diagram()  # always update

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Event handler for mouse move events.
        """

        pos = self.mapToScene(event.pos())
        self.refresh(pos=pos)

    def update_position_at_the_diagram(self) -> None:
        """
        Updates the element position in the diagram (to save)
        :return:
        """
        lat, long = self.editor.to_lat_lon(self.rect().x(), self.rect().y())

        self.lat = lat
        self.lon = long

        self.editor.update_diagram_element(device=self.api_object,
                                           latitude=lat,
                                           longitude=long,
                                           graphic_object=self)

    def refresh(self, pos: QPointF):
        """
        This function refreshes the x, y coordinates of the graphic item
        Returns:

        """

        x = pos.x() - self.rect().width() / 2
        y = pos.y() - self.rect().height() / 2
        self.setRect(x, y, self.rect().width(), self.rect().height())
        self.set_callbacks(pos.x(), pos.y())


        self.update_position_at_the_diagram()  # always update

    def set_size(self, r: float):
        """

        :param r: radius in pixels
        :return:
        """
        if r != self.size:
            rect = self.rect()
            rect.setWidth(r)
            rect.setHeight(r)

            # change the width and height while keeping the same center
            r2 = (self.size - r) / 2
            new_x = rect.x() + r2
            new_y = rect.y() + r2

            self.size = r

            # Set the new rectangle with the updated dimensions
            self.setRect(new_x, new_y, r, r)

    def recolour_mode(self):
        """
        Change the colour according to the system theme
        """
        super().recolour_mode()

        return self.set_api_object_color()

    def delete(self):
        """
        Remove this element
        @return:
        """

        deleted, delete_from_db_final = self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)

    def plot(self):
        """
        Plot API objects profiles
        """
        # time series object from the last simulation
        ts = self.editor.circuit.time_profile

        # plot the profiles
        self.api_object.plot_profiles(time=ts)


    def mouseDoubleClickEvent(self, event, /):
        """

        :param event:
        :return:
        """
        super().mouseDoubleClickEvent(event)
        self.set_api_object_color()

    def get_base_context_menu(self) -> QMenu:
        """
        Generate the base menu for injections
        :return:
        """
        menu = QMenu()

        add_menu_entry(menu=menu,
                       text="Plot profiles",
                       function_ptr=self.plot,
                       icon_path=":/Icons/icons/plot.png")

        add_menu_entry(menu=menu,
                       text="Consolidate coordinates",
                       function_ptr=self.editor.consolidate_object_coordinates,
                       icon_path=":/Icons/icons/assign_to_profile.png")

        menu.addSeparator()

        add_menu_entry(menu=menu,
                       text="Delete",
                       function_ptr=self.delete,
                       icon_path=":/Icons/icons/delete_schematic.png")

        return menu

    def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = self.get_base_context_menu()
        menu.exec(event.screenPos())

    def consolidate_coordinates(self):
        """
        Consolidate coordinates in to the DB
        """
        lat, long = self.editor.to_lat_lon(self.rect().x(), self.rect().y())
        self.api_object.latitude = lat
        self.api_object.longitude = long
        self.editor.gui.show_info_toast("Coordinates consolidated!")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import List, Dict, Union
from warnings import warn

from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGridEngine.enumerations import DeviceType
from VeraGrid.Gui.Diagrams.SchematicWidget.Substation.bus_graphics import BusGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_node_graphics import FluidNodeGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_path_graphics import FluidPathGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics import LineGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.winding_graphics import WindingGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.dc_line_graphics import DcLineGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.transformer2w_graphics import TransformerGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.hvdc_graphics import HvdcGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.vsc_graphics import VscGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.upfc_graphics import UpfcGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.series_reactance_graphics import SeriesReactanceGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.transformer3w_graphics import Transformer3WGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.generator_graphics import GeneratorGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Branches.map_line_container import MapLineContainer
from VeraGrid.Gui.Diagrams.MapWidget.Substation.voltage_level_graphic_item import VoltageLevelGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Branches.line_location_graphic_item import LineLocationGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Substation.substation_graphic_item import SubstationGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Branches.map_line_segment import MapLineSegment
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_battery_graphics import MapBatteryGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_generator_graphics import MapGeneratorGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_load_graphics import MapLoadGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_external_grid_graphics import MapExternalGridGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_static_generator_graphics import MapStaticGeneratorGraphicItem

ALL_BUS_BRACH_GRAPHICS = Union[
    BusGraphicItem,
    FluidNodeGraphicItem,
    FluidPathGraphicItem,
    LineGraphicItem,
    WindingGraphicItem,
    DcLineGraphicItem,
    TransformerGraphicItem,
    HvdcGraphicItem,
    VscGraphicItem,
    UpfcGraphicItem,
    SeriesReactanceGraphicItem,
    LineGraphicTemplateItem,
    Transformer3WGraphicItem,
    GeneratorGraphicItem,
]

ALL_MAP_GRAPHICS = Union[
    MapLineContainer,
    VoltageLevelGraphicItem,
    LineLocationGraphicItem,
    SubstationGraphicItem,
    MapLineSegment,
    MapBatteryGraphicItem,
    MapGeneratorGraphicItem,
    MapLoadGraphicItem,
    MapExternalGridGraphicItem,
    MapStaticGeneratorGraphicItem
]

ALL_GRAPHICS = Union[ALL_BUS_BRACH_GRAPHICS, ALL_MAP_GRAPHICS]


class GraphicsManager:
    """
    Class to handle the correspondance between graphics and database devices
    """

    def __init__(self) -> None:
        # this is a dictionary that groups by 2 levels:
        # first by DeviceType
        # second idtag -> GraphicItem
        self.graphic_dict: Dict[DeviceType, Dict[str, ALL_GRAPHICS]] = dict()

    def clear(self):
        """
        Clear all graphics references
        """
        self.graphic_dict.clear()

    def add_device(self, elm: ALL_DEV_TYPES, graphic: ALL_GRAPHICS) -> None:
        """
        Add the graphic of a device
        :param elm: Any database device
        :param graphic: Corresponding graphic
        """
        if graphic is not None:  # it makes no sense to add a None graphic

            elm_dict: Dict[str, ALL_GRAPHICS] = self.graphic_dict.get(elm.device_type, None)

            if elm_dict is None:
                # the graphic category did not exist
                self.graphic_dict[elm.device_type] = {elm.idtag: graphic}
            else:
                # The category exits, try to get the existing element
                graphic_0 = elm_dict.get(elm.idtag, None)
                if graphic_0 is None:
                    # the element did not exist, add it
                    elm_dict[elm.idtag] = graphic
                else:
                    # the element existed, and maybe we're going to replace it with something else
                    if graphic_0 != graphic:
                        print(f"Replacing {graphic} with {graphic}, this could be a sign of an idtag bug")
                    elm_dict[elm.idtag] = graphic

        else:
            raise ValueError(f"Trying to set a None graphic object for {elm}")

    def delete_device(self, device: ALL_DEV_TYPES) -> Union[ALL_GRAPHICS, None]:
        """
        Delete device from the registry and return the object if it exists
        :param device: Any database device
        :return: Corresponding graphic or None
        """
        if device is not None:
            # check if the category exists ...
            elm_dict = self.graphic_dict.get(device.device_type, None)

            if elm_dict is not None:
                # the category does exist, delete_with_dialogue from it
                graphic = elm_dict.get(device.idtag, None)

                if graphic:
                    del elm_dict[device.idtag]
                    return graphic

            else:
                # not found so we're ok
                return None
        else:
            return None

    def query(self, elm: ALL_DEV_TYPES) -> Union[None, ALL_GRAPHICS]:
        """
        Query the graphic of a database element
        :param elm: Any database element
        :return: Corresponding graphic
        """
        if elm is None:
            return None

        else:
            elm_dict: Dict[str, ALL_GRAPHICS] = self.graphic_dict.get(elm.device_type, None)

            if elm_dict is None:
                return None
            else:
                return elm_dict.get(elm.idtag, None)

    def query_preferring_busbars(self, elm: ALL_DEV_TYPES) -> Union[None, ALL_GRAPHICS]:
        """
        Because some connectivity nodes are graphically substituted by BusBars, we need to do this
        :param elm: Any device
        :return: Any graphic element if found, None otherwise
        """
        return self.query(elm=elm)

    def get_device_type_list(self, device_type: DeviceType) -> List[ALL_GRAPHICS]:
        """
        Get the list of graphics of a device type
        :param device_type: DeviceType
        :return: List[ALL_GRAPHICS]
        """
        elm_dict: Dict[str, ALL_GRAPHICS] = self.graphic_dict.get(device_type, None)

        if elm_dict is None:
            return list()
        else:
            return [graphic for idtag, graphic in elm_dict.items()]

    def get_device_type_dict(self, device_type: DeviceType) -> Dict[str, ALL_GRAPHICS]:
        """
        Get the list of graphics of a device type
        :param device_type: DeviceType
        :return: Dict[str, ALL_GRAPHICS]
        """
        return self.graphic_dict.get(device_type, dict())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from PySide6.QtGui import QPalette, QColor


def css_rgb(color, a=False):
    """Get a CSS `rgb` or `rgba` string from a `QtGui.QColor`."""
    return ("rgba({}, {}, {}, {})" if a else "rgb({}, {}, {})").format(*color.getRgb())


class QDarkPalette(QPalette):
    """Dark palette for a Qt application meant to be used with the Fusion theme."""
    def __init__(self, *__args):
        super().__init__(*__args)

        self.WHITE = QColor(255, 255, 255)
        self.BLACK = QColor(0, 0, 0)
        self.RED = QColor(255, 0, 0)
        self.PRIMARY = QColor(53, 53, 53)
        self.SECONDARY = QColor(35, 35, 35)
        self.TERTIARY = QColor(42, 130, 218)

        # Set all the colors based on the constants in globals
        self.setColor(QPalette.Window,          self.PRIMARY)
        self.setColor(QPalette.WindowText,      self.WHITE)
        self.setColor(QPalette.Base,            self.SECONDARY)
        self.setColor(QPalette.AlternateBase,   self.PRIMARY)
        self.setColor(QPalette.ToolTipBase,     self.WHITE)
        self.setColor(QPalette.ToolTipText,     self.WHITE)
        self.setColor(QPalette.Text,            self.WHITE)
        self.setColor(QPalette.Button,          self.PRIMARY)
        self.setColor(QPalette.ButtonText,      self.WHITE)
        self.setColor(QPalette.BrightText,      self.RED)
        self.setColor(QPalette.Link,            self.TERTIARY)
        self.setColor(QPalette.Highlight,       self.TERTIARY)
        self.setColor(QPalette.HighlightedText, self.BLACK)

    def set_stylesheet(self, app):
        """
        Static method to set the tooltip stylesheet to a `QtWidgets.QApplication`
        """
        app.setStyleSheet("QToolTip {{"
                          "color: {white};"
                          "background-color: {tertiary};"
                          "border: 1px solid {white};"
                          "}}".format(white=css_rgb(self.WHITE), tertiary=css_rgb(self.TERTIARY)))

    def set_app(self, app):
        """
        Set the Fusion theme and this palette to a `QtWidgets.QApplication`.
        """
        app.setStyle("Fusion")
        app.setPalette(self)
        self.set_stylesheet(app)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'sigma_analysis_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QHBoxLayout, QHeaderView,
    QMainWindow, QMenu, QMenuBar, QSizePolicy,
    QSplitter, QStatusBar, QTableView, QVBoxLayout,
    QWidget)

from VeraGrid.Gui.Widgets.matplotlibwidget import MatplotlibWidget
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(841, 518)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/sigma.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.actionCopy_to_clipboard = QAction(MainWindow)
        self.actionCopy_to_clipboard.setObjectName(u"actionCopy_to_clipboard")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/copy.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionCopy_to_clipboard.setIcon(icon1)
        self.actionSave = QAction(MainWindow)
        self.actionSave.setObjectName(u"actionSave")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/import_profiles.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSave.setIcon(icon2)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.splitter_3 = QSplitter(self.centralwidget)
        self.splitter_3.setObjectName(u"splitter_3")
        self.splitter_3.setOrientation(Qt.Horizontal)
        self.frame_8 = QFrame(self.splitter_3)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setFrameShape(QFrame.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.tableView = QTableView(self.frame_8)
        self.tableView.setObjectName(u"tableView")

        self.verticalLayout_7.addWidget(self.tableView)

        self.splitter_3.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.splitter_3)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName(u"plotwidget")

        self.horizontalLayout.addWidget(self.plotwidget)

        self.splitter_3.addWidget(self.PlotFrame)

        self.verticalLayout.addWidget(self.splitter_3)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 841, 22))
        self.menuActions = QMenu(self.menubar)
        self.menuActions.setObjectName(u"menuActions")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuActions.menuAction())
        self.menuActions.addAction(self.actionCopy_to_clipboard)
        self.menuActions.addAction(self.actionSave)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.actionCopy_to_clipboard.setText(QCoreApplication.translate("MainWindow", u"Copy to clipboard", None))
        self.actionSave.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.menuActions.setTitle(QCoreApplication.translate("MainWindow", u"Actions", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import os
import time
from math import isclose
from typing import List
from PySide6.QtCore import QThread, Signal
from PySide6 import QtGui
from PySide6.QtCore import Qt

from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import SyncIssueType
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.IO.file_handler import FileOpen
from VeraGridEngine.Devices.Parents.editable_device import EditableDevice, DeviceType


class SyncIssue:
    """
    SyncIssue
    """

    def __init__(self, device_type, issue_type: SyncIssueType, property_name, my_elm, their_elm):
        """

        :param device_type:
        :param issue_type:
        :param property_name:
        :param my_elm:
        :param their_elm:
        """

        self.device_type = device_type

        self.issue_type = issue_type

        self.property_name = property_name

        self.my_elm = my_elm

        self.their_elm = their_elm

        self.__accept__ = True

    def __str__(self):
        if self.issue_type == SyncIssueType.Conflict:
            return 'Mod::' + self.device_type.value + ', ' + self.issue_type.value + ', ' + self.my_elm.name + ':' + self.property_name
        elif self.issue_type == SyncIssueType.Added:
            return 'Add::' + self.device_type.value + ', ' + self.issue_type.value + ', ' + self.their_elm.name
        elif self.issue_type == SyncIssueType.Deleted:
            return 'Del::' + self.device_type.value + ', ' + self.issue_type.value + ', ' + self.my_elm.name
        else:
            return ""

    def accept(self):
        self.__accept__ = True

    def reject(self):
        self.__accept__ = False

    def accepted(self):
        return self.__accept__

    def get_my_name(self):
        return str(self.my_elm)

    def get_my_value(self):
        if self.my_elm is not None and self.property_name != "":
            return getattr(self.my_elm, self.property_name)
        else:
            return ""

    def get_their_value(self):
        if self.their_elm is not None and self.property_name != "":
            return getattr(self.their_elm, self.property_name)
        else:
            return ""

    def accept_change(self):
        """
        process the change
        :return:
        """
        their_val = self.get_their_value()
        setattr(self.my_elm, self.property_name, their_val)


def compare_devices(dev1: EditableDevice, dev2: EditableDevice):
    """
    Compare two devices and return the list of differences
    :param dev1: Device 1
    :param dev2: Device 2
    :return: list of differences [property name, value1, value2]
    """

    differences = list()
    if dev1.device_type == dev2.device_type:
        for prop_name, value in dev1.registered_properties.items():

            val1 = getattr(dev1, prop_name)
            val2 = getattr(dev2, prop_name)

            if value.tpe in [int, float]:
                if not isclose(val1, val2, abs_tol=1e-6):
                    differences.append([prop_name, val1, val2])
            else:
                if str(val1) != str(val2):
                    differences.append([prop_name, val1, val2])

    return differences


def compare_devices_lists(dev_list1, dev_list2):
    """
    Compare two devices lists
    :param dev_list1: list of devices 1
    :param dev_list2: list of devices 2
    :return: List of issues
    """

    items_dict1 = {elm.name: elm for elm in dev_list1}
    items_dict2 = {elm.name: elm for elm in dev_list2}

    issues = list()  # device type, issue type, property, my value, their value, elm1, elm2

    # check my buses against the file buses
    for name1, elm1 in items_dict1.items():

        if name1 in items_dict2.keys():

            # get the other element
            elm2 = items_dict2[name1]

            # perform the comparison in the same types, compare
            ls = compare_devices(elm1, elm2)
            for prop, val1, val2 in ls:
                issue = SyncIssue(device_type=elm1.device_type,
                                  issue_type=SyncIssueType.Conflict,
                                  property_name=prop,
                                  my_elm=elm1,
                                  their_elm=elm2)
                issues.append(issue)

            # if this is a bus, then examine the children
            if elm1.device_type == DeviceType.BusDevice:

                issues += compare_devices_lists(elm1.loads, elm2.loads)
                issues += compare_devices_lists(elm1.generators, elm2.generators)
                issues += compare_devices_lists(elm1.batteries, elm2.batteries)
                issues += compare_devices_lists(elm1.static_generators, elm2.static_generators)
                issues += compare_devices_lists(elm1.shunts, elm2.shunts)

        else:
            # my element has been deleted
            issue = SyncIssue(device_type=elm1.device_type,
                              issue_type=SyncIssueType.Deleted,
                              property_name="",
                              my_elm=elm1,
                              their_elm=None)
            issues.append(issue)

    # check the file buses against mine, here we only need to check for disagreements in the existence
    for name2, elm2 in items_dict2.items():
        if name2 not in items_dict1.keys():
            # new element added
            issue = SyncIssue(device_type=elm2.device_type,
                              issue_type=SyncIssueType.Added,
                              property_name="",
                              my_elm=None,
                              their_elm=elm2)
            issues.append(issue)

    return issues


def detect_changes_and_conflicts(current_circuit: MultiCircuit, file_circuit: MultiCircuit):
    """
    Detect changes
    :param current_circuit:
    :param file_circuit:
    :return:
    """

    issues = list()  # device name, issue type, property, my value, their value, elm1, elm2

    # buses
    issues += compare_devices_lists(dev_list1=current_circuit.buses,
                                    dev_list2=file_circuit.buses)

    # Branches
    issues += compare_devices_lists(dev_list1=current_circuit.lines,
                                    dev_list2=file_circuit.lines)

    issues += compare_devices_lists(dev_list1=current_circuit.transformers2w,
                                    dev_list2=file_circuit.transformers2w)

    issues += compare_devices_lists(dev_list1=current_circuit.hvdc_lines,
                                    dev_list2=file_circuit.hvdc_lines)

    return issues


def model_check(current_circuit: MultiCircuit, file_circuit: MultiCircuit):
    """
    Perform model check
    :param current_circuit:
    :param file_circuit:
    :return: list of issues
    """

    issues = detect_changes_and_conflicts(current_circuit, file_circuit)

    if (current_circuit.model_version + 1) <= file_circuit.model_version:
        # the remote file version is newer.
        # The first action is to attempt to merge, looking out for conflict
        # to merge, we incorporate the remote file changes and then we save

        version_conflict = True

    else:
        # there is no version conflict, so we just save
        version_conflict = False

    return issues, version_conflict


def get_issues_tree_view_model(issues: List[SyncIssue]):
    """
    Get TreeView model of the issues
    :param issues: list of issues
    :return: Model for a TreeView
    """
    # structure the issues by issue type and by device type
    k = 0
    data = dict()
    for issue in issues:

        # device_type
        # issue_type
        # prop
        # val1
        # val2
        # elm1
        # elm2

        if issue.issue_type in data.keys():

            if issue.device_type.value in data[issue.issue_type.value].keys():
                data[issue.issue_type.value][issue.device_type.value].append((k, issue))
            else:
                data[issue.issue_type.value][issue.device_type.value] = [(k, issue)]

        else:
            data[issue.issue_type.value] = {issue.device_type.value: [(k, issue)]}

        k += 1

    # build the tree
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels(['Issue #', 'name', 'property', 'my value', 'their value', 'Accept theirs'])

    # populate data
    for issue_name, devices in data.items():

        parent1 = QtGui.QStandardItem(issue_name)  # add the issue group

        for device, list_of_issues in devices.items():

            parent2 = QtGui.QStandardItem(device)  # add the device type group

            for k, issue_item in list_of_issues:  # add the row of information

                items = list()
                items.append(QtGui.QStandardItem(str(k)))
                items.append(QtGui.QStandardItem(issue_item.get_my_name()))
                items.append(QtGui.QStandardItem(issue_item.property_name))
                items.append(QtGui.QStandardItem(str(issue_item.get_my_value())))
                items.append(QtGui.QStandardItem(str(issue_item.get_their_value())))

                check = QtGui.QStandardItem(str(issue_item.__accept__))
                check.isCheckable()

                if issue_item.__accept__:
                    check.setCheckState(Qt.Checked)
                else:
                    check.setCheckState(Qt.Unchecked)

                items.append(check)

                for item in items:
                    item.setEditable(False)
                parent2.appendRow(items)

            parent2.setEditable(False)
            parent1.appendRow(parent2)

        parent1.setEditable(False)
        model.appendRow(parent1)

    return model


class FileSyncThread(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    sync_event = Signal()
    items_processed_event = Signal()

    def __init__(self, circuit: MultiCircuit, file_name, sleep_time):
        """

        :param circuit: current circuit
        :param file_name: name of the file to sync
        :param sleep_time: seconds between executions
        """
        QThread.__init__(self)

        self.circuit = circuit

        self.file_name = file_name

        self.sleep_time = sleep_time

        self.issues = list()

        self.version_conflict = False

        self.highest_version = 0

        self.logger = Logger()

        self.error_msg = ''

        self.__cancel__ = False

        self.__pause__ = False

    def run(self):
        """
        run the file save procedure
        """

        # create class to open the file
        fopen = FileOpen(self.file_name)

        while not self.__cancel__:

            if not self.__pause__:

                if os.path.exists(self.file_name):

                    # load the remote file
                    file_circuit = fopen.open(text_func=self.progress_text.emit,
                                              progress_func=self.progress_signal.emit)

                    if file_circuit is not None:
                        # sync the models
                        self.issues, self.version_conflict = model_check(self.circuit, file_circuit)

                        self.highest_version = max(self.circuit.model_version, file_circuit.model_version)

                        # notify the external world that we did sync
                        self.sync_event.emit()
                    else:
                        # the sync failed because the file was being used by another sync process
                        pass

                else:
                    # the file disappeared!
                    self.logger.add('File missing for synchronization:' + self.file_name)
                    self.cancel()

                # sleep
                time.sleep(self.sleep_time)

            else:
                # sleep 1 second to catch other events
                time.sleep(1)

        # post events
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the sync checking
        """
        self.__cancel__ = True

    def pause(self):
        """
        Pause the sync checking
        """
        self.__pause__ = True

    def resume(self):
        """
        Resume the sync checking
        """
        self.__pause__ = False

    def process_issues(self):
        """
        Process all the issues
        :return:
        """

        self.items_processed_event.emit()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import numpy as np
import rlcompleter
from PySide6.QtGui import QFont, QFontMetrics, Qt
from PySide6 import QtWidgets, QtCore
from PySide6.QtWidgets import QCompleter, QPlainTextEdit
from PySide6.QtCore import Qt, QStringListModel
from PySide6.QtGui import QTextCursor

from VeraGridEngine.IO.file_system import scripts_path
from VeraGrid.Gui.Main.SubClasses.io import IoMain
from VeraGrid.Gui.python_highlighter import PythonHighlighter
from VeraGrid.Gui.gui_functions import CustomFileSystemModel
import VeraGrid.Gui.gui_functions as gf
from VeraGrid.Gui.messages import error_msg, yes_no_question
from VeraGrid.Gui.font_config import CONSOLE_TEXT_SIZE


class ScriptingMain(IoMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        IoMain.__init__(self, parent)

        self.add_console_vars()

        # scripts tree view --------------------------------------------------------------------------------------------
        self.python_fs_model = CustomFileSystemModel(root_path=scripts_path(), ext_filter=['*.py'])
        self.ui.sourceCodeTreeView.setModel(self.python_fs_model)
        self.ui.sourceCodeTreeView.setRootIndex(self.python_fs_model.index(scripts_path()))

        # actions ------------------------------------------------------------------------------------------------------
        self.ui.actionReset_console.triggered.connect(self.reset_console)

        # button clicks ------------------------------------------------------------------------------------------------
        self.ui.runSourceCodeButton.clicked.connect(self.run_source_code)
        self.ui.saveSourceCodeButton.clicked.connect(self.save_source_code)
        self.ui.clearSourceCodeButton.clicked.connect(self.clear_source_code)
        self.ui.clearConsoleButton.clicked.connect(self.reset_console)

        # double clicked -----------------------------------------------------------------------------------------------
        self.ui.sourceCodeTreeView.doubleClicked.connect(self.source_code_tree_clicked)

        # context menu
        self.ui.sourceCodeTreeView.customContextMenuRequested.connect(self.show_source_code_tree_context_menu)

        # Set context menu policy to CustomContextMenu
        self.ui.sourceCodeTreeView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

    # def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent):
    #     """
    #     Event filter to capture Ctrl + Enter
    #     :param watched:
    #     :param event:
    #     :return:
    #     """
    #
    #     if watched == self.ui.sourceCodeTextEdit and event.type() == QtCore.QEvent.Type.KeyPress:
    #         if event.key() == Qt.Key.Key_Return and event.modifiers() & Qt.KeyboardModifier.ControlModifier:
    #             self.run_source_code()
    #             return True
    #
    #     return super().eventFilter(watched, event)

    def append_output(self, text: str):
        """
        Add some text to the output
        :param text: text to append
        """
        self.console.append_output(text)

    def run_source_code(self):
        """
        Run the source code in the IPython console
        """
        source_code = self.code_editor.toPlainText()

        if source_code[-1] != '\n':
            source_code += "\n"

        self.console.execute(source_code)
        self.console.append_output(">>> ")

    def source_code_tree_clicked(self, index):
        """
        On double click on a source code tree item, load the source code
        """
        pth = self.python_fs_model.filePath(index)

        if os.path.exists(pth):
            with open(pth, 'r') as f:
                txt = "\n".join(line.rstrip() for line in f)
                self.code_editor.setPlainText(txt)

            name = os.path.basename(pth)
            self.ui.sourceCodeNameLineEdit.setText(name.replace('.py', ''))
        else:
            error_msg(pth + ' does not exists :/', 'Open script')

    def clear_source_code(self):
        """
        Clear source code
        """
        ok = yes_no_question(text='Are you sure you want to clear source code?',
                             title='Clear Source Code')

        if ok:
            self.ui.sourceCodeNameLineEdit.setText("")
            self.code_editor.setPlainText("")

    def save_source_code(self):
        """
        Save the source code
        """
        name = self.ui.sourceCodeNameLineEdit.text().strip()

        if name != '':
            fname = name + '.py'
            pth = os.path.join(scripts_path(), fname)
            with open(pth, 'w') as f:
                f.write(self.code_editor.toPlainText())
        else:
            error_msg("Please enter a name for the script", title="Save script")

    def delete_source_code(self):
        """
        Delete the selected file
        """
        index = self.ui.sourceCodeTreeView.currentIndex()
        pth = self.python_fs_model.filePath(index)
        if os.path.exists(pth):
            ok = yes_no_question(text="Do you want to delete_with_dialogue {}?".format(pth),
                                 title="Delete source code file")

            if ok:
                os.remove(pth)
        else:
            error_msg(pth + ' does not exists :/', "Delete source code file")

    def show_source_code_tree_context_menu(self, pos: QtCore.QPoint):
        """
        Show source code tree view context menu
        :param pos: Relative click position
        """
        context_menu = QtWidgets.QMenu(parent=self.ui.diagramsListView)

        gf.add_menu_entry(menu=context_menu,
                          text="Delete",
                          icon_path=":/Icons/icons/delete_with_dialogue.png",
                          function_ptr=self.delete_source_code)

        # Convert global position to local position of the list widget
        mapped_pos = self.ui.sourceCodeTreeView.viewport().mapToGlobal(pos)
        context_menu.exec(mapped_pos)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Dict, List
import numpy as np
import pandas as pd


from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton,
    QComboBox, QListWidget, QDialogButtonBox, QMenu
)
from PySide6.QtCore import Qt, QPoint

from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Simulations.Rms.rms_results import RmsResults, ResultsTable
from VeraGridEngine.enumerations import DeviceType
import numpy as np
import pandas as pd


class RmsPlotDialog(QDialog):
    """
    Special plot for dynamic variables
    """

    def __init__(self, grid: MultiCircuit, results: RmsResults, parent=None):
        super().__init__(parent)

        # --- Preparar dispositivos y variables ---
        buses = [bus for bus in grid.get_buses()]
        branches = [branch for branch in grid.get_branches_iter()]
        injections = [injection for injection in grid.get_injection_devices_iter()]
        devices = buses + branches + injections

        devices_options = {}
        for device in devices:
            devices_options[device.name] = [
                results.uid2vars_glob_name[var.uid]
                for var in (
                    device.rms_model.model.state_vars +
                    device.rms_model.model.algebraic_vars
                )
            ]

        self.setWindowTitle("Plot Variables")
        self.uid2idx = results.uid2idx
        self.vars_glob_name2uid = results.vars_glob_name2uid
        self.devices = devices_options

        # --- ResultsTable ---
        self.results_table = ResultsTable(
            data=np.array(results.values),
            index=np.array(pd.to_datetime(results.time_array).astype(str), dtype=np.str_),
            columns=results.variable_array,
            title="Rms Simulation Results",
            units=results.units,
            idx_device_type=DeviceType.TimeDevice,
            cols_device_type=DeviceType.NoDevice,
            xlabel="time (s)",
            ylabel="",
        )

        self.selected_vars = []

        # --- Layout principal ---
        layout = QVBoxLayout(self)

        # --- Selector de dispositivo ---
        dev_layout = QHBoxLayout()
        dev_layout.addWidget(QLabel("Device:"))
        self.device_combo = QComboBox()
        self.device_combo.addItems(list(devices_options.keys()))
        self.device_combo.currentIndexChanged.connect(self.update_variables)
        dev_layout.addWidget(self.device_combo)
        layout.addLayout(dev_layout)

        # --- Selector de variable ---
        var_layout = QHBoxLayout()
        var_layout.addWidget(QLabel("Variable:"))
        self.var_combo = QComboBox()
        var_layout.addWidget(self.var_combo)
        layout.addLayout(var_layout)

        # --- Botón para añadir variable ---
        add_btn = QPushButton("Add")
        add_btn.clicked.connect(self.add_variable)
        layout.addWidget(add_btn)

        # --- Lista de variables seleccionadas ---
        self.list_widget = QListWidget()
        self.list_widget.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.list_widget.customContextMenuRequested.connect(self.show_variable_context_menu)
        layout.addWidget(self.list_widget)

        # --- Canvas embebido ---
        self.figure = Figure(figsize=(6, 3))
        self.ax = self.figure.add_subplot(111)
        self.canvas = FigureCanvas(self.figure)
        layout.addWidget(self.canvas)

        # --- Botones inferiores ---
        buttons_layout = QHBoxLayout()

        self.buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttons.accepted.connect(self.plot_selected)
        self.buttons.rejected.connect(self.reject)
        buttons_layout.addWidget(self.buttons)

        # --- Botón ventana separada ---
        show_window_btn = QPushButton("Show in new window")
        show_window_btn.clicked.connect(self.show_external_plot)
        buttons_layout.addWidget(show_window_btn)

        layout.addLayout(buttons_layout)

        # --- Inicializar variables ---
        self.update_variables(0)

    # -------------------------------------------------------------

    def update_variables(self, index):

        device = self.device_combo.currentText()
        self.var_combo.clear()
        self.var_combo.addItems(self.devices[device])

    # -------------------------------------------------------------

    def add_variable(self):

        var = self.var_combo.currentText()
        if var and var not in [self.list_widget.item(i).text() for i in range(self.list_widget.count())]:
            self.selected_vars.append(self.vars_glob_name2uid[var])
            self.list_widget.addItem(var)
            self.plot_selected()

    # -------------------------------------------------------------

    def show_variable_context_menu(self, pos: QPoint):

        item = self.list_widget.itemAt(pos)
        if item is not None:
            menu = QMenu(self)
            remove_action = menu.addAction("Remove variable")
            action = menu.exec(self.list_widget.mapToGlobal(pos))
            if action == remove_action:
                self.remove_variable(item)

    # -------------------------------------------------------------

    def remove_variable(self, item):

        var_name = item.text()
        if var_name in self.vars_glob_name2uid:
            uid_to_remove = self.vars_glob_name2uid[var_name]
            if uid_to_remove in self.selected_vars:
                self.selected_vars.remove(uid_to_remove)

        row = self.list_widget.row(item)
        self.list_widget.takeItem(row)
        self.plot_selected()

    # -------------------------------------------------------------

    def plot_selected(self):

        self.ax.clear()
        if not self.selected_vars:
            self.canvas.draw()
            return

        selected_col_idx = [self.uid2idx[uid] for uid in self.selected_vars]
        self.results_table.plot(ax=self.ax, selected_col_idx=selected_col_idx)
        self.canvas.draw()

    # -------------------------------------------------------------

    def show_external_plot(self):
        if not self.selected_vars:
            return

        selected_col_idx = [self.uid2idx[uid] for uid in self.selected_vars]

        # Create a separate dialog for the plot
        external_dialog = QDialog(self)
        external_dialog.setWindowTitle("Plot Window")
        external_dialog.resize(900, 500)

        layout = QVBoxLayout(external_dialog)

        # Create figure and canvas
        figure = Figure(figsize=(10, 5))
        ax = figure.add_subplot(111)
        canvas = FigureCanvas(figure)

        # Add Matplotlib toolbar (includes Save, Zoom, Pan, etc.)
        toolbar = NavigationToolbar(canvas, external_dialog)
        layout.addWidget(toolbar)

        layout.addWidget(canvas)

        # Plot data
        self.results_table.plot(ax=ax, selected_col_idx=selected_col_idx)
        canvas.draw()

        external_dialog.show()






# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import sys
import pandas as pd
from PySide6 import QtWidgets

from VeraGrid.Gui.LoadCatalogue.SelectComponents import Ui_MainWindow
from VeraGridEngine.Devices import TransformerType, UndergroundLineType, SequenceLineType
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
import VeraGrid.Session.synchronization_driver as syncdrv


class CatalogueGUI(QtWidgets.QDialog):

    def __init__(self, parent=None):
        """
        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.selected_file = None
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('Custom Catalogue')
        self.circuit: MultiCircuit = MultiCircuit()

        self.ui.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.ui.buttonBox.accepted.connect(self.on_accept)
        self.ui.buttonBox.rejected.connect(self.on_reject)

        self.file_sync_thread = syncdrv.FileSyncThread(self.circuit, None, None)

    def on_accept(self):
        if self.ui.checkBox.isChecked() or self.ui.checkBox_2.isChecked() or self.ui.checkBox_3.isChecked():
            self.open_file_threaded()
            self.accept()
        else:
            quit_msg = "No checkbox was selected for adding the component to the catalogue."
            QtWidgets.QMessageBox.warning(self, "Can't upload file", quit_msg)
            print("No checkbox was selected for adding the component to the catalogue.")

    def on_reject(self):
        self.reject()

    def read_csv_file(self, filename):
        return pd.read_csv(filename)

    def open_file_threaded(self, post_function=None, title: str = 'Open file'):
        files_types = "CSV (*.csv)"
        dialogue = QtWidgets.QFileDialog(None, caption=title, filter=f"Formats ({files_types})")
        if dialogue.exec():
            self.selected_file = dialogue.selectedFiles()[0]
            if post_function:
                post_function()

    def get_transformer_catalogue(self, df):
        lst = list()
        for _, item in df.iterrows():
            transformer = TransformerType(hv_nominal_voltage=item['HV (kV)'],
                                          lv_nominal_voltage=item['LV (kV)'],
                                          nominal_power=item['Rate (MVA)'],
                                          copper_losses=item['Copper losses (kW)'],
                                          iron_losses=item['No load losses (kW)'],
                                          no_load_current=item['No load current (%)'],
                                          short_circuit_voltage=item['V short circuit (%)'],
                                          gr_hv1=0.5,
                                          gx_hv1=0.5,
                                          name=item['Name'])
            lst.append(transformer)
        return lst

    def get_cables_catalogue(self, df):
        lst = list()
        for _, item in df.iterrows():
            tpe = UndergroundLineType(name=item['Name'],
                                      Imax=item['Rated current [kA]'],
                                      Vnom=item['Rated voltage [kV]'],
                                      R=item['R [Ohm/km AC@20°C]'],
                                      X=item['X [Ohm/km]'],
                                      B=0.0,
                                      R0=item['R0 (AC) [Ohm/km]'],
                                      X0=item['X0  [Ohm/km]'],
                                      B0=0.0)
            lst.append(tpe)

        return lst

    def get_sequence_lines_catalogue(self, df):
        lst = list()
        for i, item in df.iterrows():
            tpe = SequenceLineType(name=item['Name'],
                                   Vnom=item['Vnom (kV)'],
                                   Imax=item['Imax (kA)'],
                                   R=item['r (ohm/km)'],
                                   X=item['x (ohm/km)'],
                                   B=item['b (uS/km)'],
                                   R0=item['r0 (ohm/km)'],
                                   X0=item['x0 (ohm/km)'],
                                   B0=item['b0 (uS/km)'])

            lst.append(tpe)
        return lst

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = CatalogueGUI()
    window.resize(int(1.61 * 400), 400)  # golden ratio
    window.show()
    sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING, Union
from PySide6.QtGui import QIcon, QPixmap
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGridEngine.Devices.Branches.vsc import VSC
from VeraGridEngine.enumerations import TapModuleControl
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class VscGraphicItem(LineGraphicTemplateItem):
    """
    Graphics item for the VSC converter
    """

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem],
                 editor: SchematicWidget,
                 width=5,
                 api_object: VSC = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """

        if from_port.parent.api_object.is_dc and not to_port.parent.api_object.is_dc:  # this is the correct sense
            # do nothing as the ports are in the correct sense
            LineGraphicTemplateItem.__init__(self=self,
                                             from_port=from_port,
                                             to_port=to_port,
                                             editor=editor,
                                             width=width,
                                             api_object=api_object,
                                             draw_labels=draw_labels)
        elif not from_port.parent.api_object.is_dc and to_port.parent.api_object.is_dc:  # opposite sense, revert
            # flip the ports
            LineGraphicTemplateItem.__init__(self=self,
                                             from_port=to_port,
                                             to_port=from_port,
                                             editor=editor,
                                             width=width,
                                             api_object=api_object,
                                             draw_labels=draw_labels)
            print("had to flip VSC graphic ports!")
        else:
            raise Exception('VscGraphics: Impossible connecting a VSC device here. '
                            'VSC devices must be connected between AC and DC buses')



    @property
    def api_object(self) -> VSC:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()

            pe = menu.addAction('Enable/Disable')
            pe_icon = QIcon()
            if self.api_object.active:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/uncheck_all.png"))
            else:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/check_all.png"))
            pe.setIcon(pe_icon)
            pe.triggered.connect(self.enable_disable_toggle)

            add_menu_entry(menu=menu,
                           text="Draw labels",
                           function_ptr=self.enable_disable_label_drawing,
                           checkeable=True,
                           checked_value=self.draw_labels)

            # rabf = menu.addAction('Change bus')
            # move_bus_icon = QIcon()
            # move_bus_icon.addPixmap(QPixmap(":/Icons/icons/move_bus.png"))
            # rabf.setIcon(move_bus_icon)
            # rabf.triggered.connect(self.change_bus)

            add_menu_entry(menu=menu,
                           text="Change bus",
                           function_ptr=self.change_bus,
                           icon_path=":/Icons/icons/move_bus.png")

            add_menu_entry(menu=menu,
                           text="Set Control dev 1",
                           function_ptr=self.set_control_dev_1,
                           icon_path=":/Icons/icons/move_bus.png")

            add_menu_entry(menu=menu,
                           text="Set Control dev 2",
                           function_ptr=self.set_control_dev_2,
                           icon_path=":/Icons/icons/move_bus.png")

            menu.addSeparator()

            # ra2 = menu.addAction('Delete')
            # del_icon = QIcon()
            # del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.png"))
            # ra2.setIcon(del_icon)
            # ra2.triggered.connect(self.delete)

            add_menu_entry(menu=menu,
                           text="Delete",
                           function_ptr=self.delete,
                           icon_path=":/Icons/icons/delete3.png")

            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Control V from",
                           function_ptr=self.control_v_from,
                           icon_path=":/Icons/icons/edit.png")

            add_menu_entry(menu=menu,
                           text="Control V to",
                           function_ptr=self.control_v_to,
                           icon_path=":/Icons/icons/edit.png")

            menu.addSeparator()

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.png"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)

            menu.exec_(event.screenPos())
        else:
            pass

    def mouseDoubleClickEvent(self, event):
        """
        On double click, edit
        :param event:
        :return:
        """

        pass

    def control_v_from(self):
        """

        :return:
        """
        self.api_object.regulation_bus = self.api_object.bus_from
        self.api_object.tap_module_control_mode = TapModuleControl.Vm

    def control_v_to(self):
        """

        :return:
        """
        self.api_object.regulation_bus = self.api_object.bus_to
        self.api_object.tap_module_control_mode = TapModuleControl.Vm

    def set_control_dev_1(self):
        """

        :return:
        """
        self.editor.set_vsc_control_dev(graphic=self, control_idx=1)

    def set_control_dev_2(self):
        """

        :return:
        """
        self.editor.set_vsc_control_dev(graphic=self, control_idx=2)

    def set_arrows_with_power_vsc(self, Pf: float | None, Pt: complex | None, Qt: complex | None) -> None:
        """
        Set the arrow directions
        :param Pf: DC actrive power from
        :param Pt: AC active power to
        :param Qt: AC reactive power
        """

        if Pf is not None:

            if Pt is None:
                Pt = -Pf

            if Qt is None:
                Qt = 0

            # from -> DC
            # to   -> AC

            # AC side
            self.arrow_p_to.set_value(Pt, True, Pt > 0, name="Pac", units="MW", draw_label=self.draw_labels)
            self.arrow_q_to.set_value(Qt, True, Qt > 0, name="Qac", units="MVAr", draw_label=self.draw_labels)

            # DC side
            self.arrow_p_from.set_value(Pf, True, Pf < 0, name="Pdc", units="MW", draw_label=self.draw_labels)
            self.arrow_q_from.setVisible(False)
        else:
            if Pt is None:
                self.arrow_p_from.setVisible(False)
                self.arrow_q_from.setVisible(False)
                self.arrow_p_to.setVisible(False)
                self.arrow_q_to.setVisible(False)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, Union
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem
from VeraGridEngine.Devices.Branches.hvdc_line import HvdcLine
from VeraGrid.Gui.messages import yes_no_question

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class HvdcGraphicItem(LineGraphicTemplateItem):

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem],
                 editor: SchematicWidget,
                 width=5,
                 api_object: HvdcLine = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """
        LineGraphicTemplateItem.__init__(self=self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         draw_labels=draw_labels)

    @property
    def api_object(self) -> HvdcLine:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("HVDC line")

            pe = menu.addAction('Active')
            pe.setCheckable(True)
            pe.setChecked(self.api_object.active)
            pe.triggered.connect(self.enable_disable_toggle)

            add_menu_entry(menu=menu,
                           text="Draw labels",
                           icon_path="",
                           function_ptr=self.enable_disable_label_drawing,
                           checkeable=True,
                           checked_value=self.draw_labels)

            # pe2 = menu.addAction('Convert to Multi-terminal')
            # pe2.triggered.connect(self.convert_to_multi_terminal)

            add_menu_entry(menu=menu,
                           text="Change bus",
                           icon_path=":/Icons/icons/move_bus.png",
                           function_ptr=self.change_bus)

            add_menu_entry(menu=menu,
                           text="Convert to VSC multi-terminal",
                           icon_path=":/Icons/icons/vsc.png",
                           function_ptr=self.convert_to_multi_terminal)

            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Plot profiles",
                           icon_path=":/Icons/icons/plot.png",
                           function_ptr=self.plot_profiles)

            add_menu_entry(menu=menu,
                           text="Assign rate to profile",
                           icon_path=":/Icons/icons/assign_to_profile.png",
                           function_ptr=self.assign_rate_to_profile)

            add_menu_entry(menu=menu,
                           text="Assign active state to profile",
                           icon_path=":/Icons/icons/assign_to_profile.png",
                           function_ptr=self.assign_status_to_profile)

            add_menu_entry(menu=menu,
                           text="Delete",
                           icon_path=":/Icons/icons/delete3.png",
                           function_ptr=self.delete)

            menu.exec_(event.screenPos())
        else:
            pass

    def convert_to_multi_terminal(self):
        """
        Convert this HvdcLine to a vsc + DC line system
        """
        ok = yes_no_question('Do you want to change the HvdcLine by 2 VSC converters + 1 DC Line?',
                             'Change by a VSC system')

        if ok:
            self.editor.convert_hvdc_line_to_vsc_system(hvdc_line=self.api_object)

    def plot_profiles(self):
        """
        Plot the time series profiles
        @return:
        """
        # get the index of this object
        i = self.editor.circuit.get_hvdc().index(self.api_object)
        self.editor.plot_hvdc_branch(i, self.api_object)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'tower_builder.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDialog, QDoubleSpinBox,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QLineEdit, QPushButton, QSizePolicy, QSpacerItem,
    QSplitter, QTableView, QVBoxLayout, QWidget)

from VeraGrid.Gui.Widgets.matplotlibwidget import MatplotlibWidget
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_TowerBuilderDialog(object):
    def setupUi(self, TowerBuilderDialog):
        if not TowerBuilderDialog.objectName():
            TowerBuilderDialog.setObjectName(u"TowerBuilderDialog")
        TowerBuilderDialog.resize(934, 714)
        font = QFont()
        font.setPointSize(9)
        TowerBuilderDialog.setFont(font)
        self.verticalLayout_3 = QVBoxLayout(TowerBuilderDialog)
        self.verticalLayout_3.setSpacing(1)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(6, 6, 6, 6)
        self.frame_6 = QFrame(TowerBuilderDialog)
        self.frame_6.setObjectName(u"frame_6")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_6.sizePolicy().hasHeightForWidth())
        self.frame_6.setSizePolicy(sizePolicy)
        self.frame_6.setMaximumSize(QSize(16777215, 32))
        self.frame_6.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Shadow.Plain)
        self.horizontalLayout_3 = QHBoxLayout(self.frame_6)
        self.horizontalLayout_3.setSpacing(5)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(1, 1, 1, 1)
        self.label_9 = QLabel(self.frame_6)
        self.label_9.setObjectName(u"label_9")

        self.horizontalLayout_3.addWidget(self.label_9)

        self.name_lineEdit = QLineEdit(self.frame_6)
        self.name_lineEdit.setObjectName(u"name_lineEdit")

        self.horizontalLayout_3.addWidget(self.name_lineEdit)

        self.label_13 = QLabel(self.frame_6)
        self.label_13.setObjectName(u"label_13")

        self.horizontalLayout_3.addWidget(self.label_13)

        self.voltage_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.voltage_doubleSpinBox.setObjectName(u"voltage_doubleSpinBox")
        self.voltage_doubleSpinBox.setDecimals(1)
        self.voltage_doubleSpinBox.setMinimum(0.100000000000000)
        self.voltage_doubleSpinBox.setMaximum(100000000.000000000000000)
        self.voltage_doubleSpinBox.setValue(100.000000000000000)

        self.horizontalLayout_3.addWidget(self.voltage_doubleSpinBox)

        self.label_8 = QLabel(self.frame_6)
        self.label_8.setObjectName(u"label_8")

        self.horizontalLayout_3.addWidget(self.label_8)

        self.frequency_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.frequency_doubleSpinBox.setObjectName(u"frequency_doubleSpinBox")
        self.frequency_doubleSpinBox.setDecimals(0)
        self.frequency_doubleSpinBox.setValue(50.000000000000000)

        self.horizontalLayout_3.addWidget(self.frequency_doubleSpinBox)

        self.label_11 = QLabel(self.frame_6)
        self.label_11.setObjectName(u"label_11")

        self.horizontalLayout_3.addWidget(self.label_11)

        self.rho_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.rho_doubleSpinBox.setObjectName(u"rho_doubleSpinBox")
        self.rho_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.rho_doubleSpinBox.setValue(100.000000000000000)

        self.horizontalLayout_3.addWidget(self.rho_doubleSpinBox)


        self.verticalLayout_3.addWidget(self.frame_6)

        self.main_splitter = QSplitter(TowerBuilderDialog)
        self.main_splitter.setObjectName(u"main_splitter")
        self.main_splitter.setOrientation(Qt.Orientation.Horizontal)
        self.frame_8 = QFrame(self.main_splitter)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame_8)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.splitter = QSplitter(self.frame_8)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.frame_3 = QFrame(self.splitter)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_8 = QVBoxLayout(self.frame_3)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.verticalLayout_8.setContentsMargins(1, 1, 1, 1)
        self.label_12 = QLabel(self.frame_3)
        self.label_12.setObjectName(u"label_12")

        self.verticalLayout_8.addWidget(self.label_12)

        self.wires_tableView = QTableView(self.frame_3)
        self.wires_tableView.setObjectName(u"wires_tableView")

        self.verticalLayout_8.addWidget(self.wires_tableView)

        self.splitter.addWidget(self.frame_3)
        self.frame_4 = QFrame(self.splitter)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_4 = QVBoxLayout(self.frame_4)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(1, 1, 1, 1)
        self.frame = QFrame(self.frame_4)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.label_10 = QLabel(self.frame)
        self.label_10.setObjectName(u"label_10")

        self.horizontalLayout_2.addWidget(self.label_10)

        self.add_to_tower_pushButton = QPushButton(self.frame)
        self.add_to_tower_pushButton.setObjectName(u"add_to_tower_pushButton")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/plus.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.add_to_tower_pushButton.setIcon(icon)

        self.horizontalLayout_2.addWidget(self.add_to_tower_pushButton)

        self.delete_from_tower_pushButton = QPushButton(self.frame)
        self.delete_from_tower_pushButton.setObjectName(u"delete_from_tower_pushButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/minus.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.delete_from_tower_pushButton.setIcon(icon1)

        self.horizontalLayout_2.addWidget(self.delete_from_tower_pushButton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)


        self.verticalLayout_4.addWidget(self.frame)

        self.tower_tableView = QTableView(self.frame_4)
        self.tower_tableView.setObjectName(u"tower_tableView")

        self.verticalLayout_4.addWidget(self.tower_tableView)

        self.splitter.addWidget(self.frame_4)
        self.frame_2 = QFrame(self.splitter)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout = QVBoxLayout(self.frame_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.frame_7 = QFrame(self.frame_2)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_7.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_4 = QHBoxLayout(self.frame_7)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(1, 1, 1, 1)
        self.label = QLabel(self.frame_7)
        self.label.setObjectName(u"label")

        self.horizontalLayout_4.addWidget(self.label)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_3)

        self.matrixViewComboBox = QComboBox(self.frame_7)
        self.matrixViewComboBox.setObjectName(u"matrixViewComboBox")

        self.horizontalLayout_4.addWidget(self.matrixViewComboBox)


        self.verticalLayout.addWidget(self.frame_7)

        self.matrixTableView = QTableView(self.frame_2)
        self.matrixTableView.setObjectName(u"matrixTableView")

        self.verticalLayout.addWidget(self.matrixTableView)

        self.splitter.addWidget(self.frame_2)

        self.verticalLayout_2.addWidget(self.splitter)

        self.main_splitter.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.main_splitter)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.Shape.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.PlotFrame)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(1, 1, 1, 1)
        self.label_4 = QLabel(self.PlotFrame)
        self.label_4.setObjectName(u"label_4")

        self.verticalLayout_7.addWidget(self.label_4)

        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName(u"plotwidget")

        self.verticalLayout_7.addWidget(self.plotwidget)

        self.frame_9 = QFrame(self.PlotFrame)
        self.frame_9.setObjectName(u"frame_9")
        self.frame_9.setMaximumSize(QSize(16777215, 24))
        self.frame_9.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_9.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_5 = QHBoxLayout(self.frame_9)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.compute_pushButton = QPushButton(self.frame_9)
        self.compute_pushButton.setObjectName(u"compute_pushButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/calc.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.compute_pushButton.setIcon(icon2)
        self.compute_pushButton.setIconSize(QSize(16, 16))

        self.horizontalLayout_5.addWidget(self.compute_pushButton)

        self.horizontalSpacer_4 = QSpacerItem(19, 19, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_4)

        self.acceptButton = QPushButton(self.frame_9)
        self.acceptButton.setObjectName(u"acceptButton")

        self.horizontalLayout_5.addWidget(self.acceptButton)


        self.verticalLayout_7.addWidget(self.frame_9)

        self.main_splitter.addWidget(self.PlotFrame)

        self.verticalLayout_3.addWidget(self.main_splitter)


        self.retranslateUi(TowerBuilderDialog)

        QMetaObject.connectSlotsByName(TowerBuilderDialog)
    # setupUi

    def retranslateUi(self, TowerBuilderDialog):
        TowerBuilderDialog.setWindowTitle(QCoreApplication.translate("TowerBuilderDialog", u"Tower creation", None))
        self.label_9.setText(QCoreApplication.translate("TowerBuilderDialog", u"Name", None))
        self.label_13.setText(QCoreApplication.translate("TowerBuilderDialog", u"Voltage", None))
        self.voltage_doubleSpinBox.setSuffix(QCoreApplication.translate("TowerBuilderDialog", u" kV", None))
        self.label_8.setText(QCoreApplication.translate("TowerBuilderDialog", u"Frequency", None))
        self.frequency_doubleSpinBox.setSuffix(QCoreApplication.translate("TowerBuilderDialog", u" Hz", None))
        self.label_11.setText(QCoreApplication.translate("TowerBuilderDialog", u"Earth resistivity", None))
        self.rho_doubleSpinBox.setSuffix(QCoreApplication.translate("TowerBuilderDialog", u" \u03a9/m^3", None))
        self.label_12.setText(QCoreApplication.translate("TowerBuilderDialog", u"Wire catalogue", None))
        self.label_10.setText(QCoreApplication.translate("TowerBuilderDialog", u"Wire composition", None))
#if QT_CONFIG(tooltip)
        self.add_to_tower_pushButton.setToolTip(QCoreApplication.translate("TowerBuilderDialog", u"Add wire", None))
#endif // QT_CONFIG(tooltip)
        self.add_to_tower_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.delete_from_tower_pushButton.setToolTip(QCoreApplication.translate("TowerBuilderDialog", u"Delete wire", None))
#endif // QT_CONFIG(tooltip)
        self.delete_from_tower_pushButton.setText("")
        self.label.setText(QCoreApplication.translate("TowerBuilderDialog", u"Matrix", None))
        self.label_4.setText(QCoreApplication.translate("TowerBuilderDialog", u"Tower", None))
#if QT_CONFIG(tooltip)
        self.compute_pushButton.setToolTip(QCoreApplication.translate("TowerBuilderDialog", u"Compute matrices", None))
#endif // QT_CONFIG(tooltip)
        self.compute_pushButton.setText("")
        self.acceptButton.setText(QCoreApplication.translate("TowerBuilderDialog", u"Accept", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import List, Set, Dict, Union, Tuple, Generator, TYPE_CHECKING
import numpy as np
import cv2
from matplotlib import pyplot as plt

from PySide6.QtCore import Qt
from PySide6.QtGui import QImage
from PySide6.QtWidgets import (QListView, QTableView, QVBoxLayout, QHBoxLayout, QFrame, QSplitter, QAbstractItemView,
                               QGraphicsItem)

from VeraGrid.Gui.Diagrams.generic_graphics import GenericDiagramWidget
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Devices.Branches.line import Line
from VeraGridEngine.Devices.Branches.dc_line import DcLine
from VeraGridEngine.Devices.Branches.hvdc_line import HvdcLine
from VeraGridEngine.Devices.Branches.transformer import Transformer2W
from VeraGridEngine.Devices.Branches.vsc import VSC
from VeraGridEngine.Devices.Branches.upfc import UPFC
from VeraGridEngine.Simulations import (PowerFlowTimeSeriesResults, LinearAnalysisTimeSeriesResults,
                                        ContingencyAnalysisTimeSeriesResults, OptimalPowerFlowTimeSeriesResults,
                                        StochasticPowerFlowResults)
from VeraGridEngine.basic_structures import Vec, CxVec, IntVec
from VeraGridEngine.Devices.Diagrams.schematic_diagram import SchematicDiagram
from VeraGridEngine.Devices.Diagrams.map_diagram import MapDiagram
from VeraGridEngine.Simulations.types import DRIVER_OBJECTS
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import SimulationTypes, ResultTypes
import VeraGridEngine.Devices.Diagrams.palettes as palettes

from VeraGrid.Gui.Diagrams.graphics_manager import GraphicsManager, ALL_GRAPHICS
from VeraGrid.Gui.general_dialogues import DeleteDialogue
from VeraGrid.Gui.messages import yes_no_question, info_msg
from VeraGrid.Gui.object_model import ObjectsModel

if TYPE_CHECKING:
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import MapLibraryModel
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicLibraryModel
    from VeraGrid.Gui.Main.SubClasses.Model.diagrams import DiagramsMain
    from VeraGrid.Gui.Main.VeraGridMain import VeraGridMainGUI


def change_font_size(obj, font_size: int):
    """

    :param obj:
    :param font_size:
    :return:
    """
    font1 = obj.font()
    font1.setPointSize(font_size)
    obj.setFont(font1)


def qimage_tocv2_by_disk(qimage: QImage, logger: Logger, file_path):
    """

    :param qimage: Qimage
    :param logger: Logger
    :param file_path: temp file path
    :return:
    """
    # Convert QImage to PNG format and save
    if not qimage.save(file_path, "PNG"):
        logger.add_error(msg=f"Error: Could not save QImage to {file_path}")
        return None

    # Use OpenCV to read the saved image
    opencv_image = cv2.imread(file_path)
    if opencv_image is None:
        logger.add_error(msg=f"Error: Could not save QImage to {file_path}")
        return None

    return opencv_image


def qimage_to_cv(qimage: QImage, logger: Logger, force_disk=False) -> np.ndarray:
    """
    Convert a image from Qt to an OpenCV image
    :param qimage: Qimage
    :param logger: Logger
    :param force_disk: if true, the image is converted by saving to disk and loading again with open-cv
    :return: OpenCv matrix
    """
    width = qimage.width()
    height = qimage.height()

    if force_disk:
        opencv_image = qimage_tocv2_by_disk(qimage, logger, file_path="__img__.png")

        return opencv_image
    else:
        try:
            # convert picture using the memory
            # we need to delete the alpha channel, otherwise the video frame is not saved
            cv_mat = np.array(qimage.constBits()).reshape(height, width, 4).astype(np.uint8)[:, :, :3]

            return cv_mat

        except ValueError as e:

            logger.add_error(msg=f"Could not convert frame: {e}, failed over to second image conversion method.")

            try:
                # Convert the QImage to RGB format if it is not already in that format
                qimage = qimage.convertToFormat(QImage.Format.Format_RGB888)

                # Get the pointer to the data and stride (bytes per line)
                ptr = qimage.bits()
                # ptr.itemsize = qimage.sizeInBytes()  # Set the size of the memoryview
                stride = qimage.bytesPerLine()  # Get the number of bytes per line (width * channels + padding)

                # Create a numpy array with the correct shape based on the stride
                arr = np.array(ptr).reshape((height, stride // 3, 3)).astype(np.uint8)  # Adjust for stride

                # Crop the width to the actual image width (in case stride > width * channels)
                arr = arr[:, :width, :]

                # Convert RGB to BGR for OpenCV
                cv_mat = cv2.cvtColor(arr, cv2.COLOR_RGB2BGR)

                return cv_mat
            except ValueError as e2:
                logger.add_error(msg=f"Could not convert frame: {e2}, failed over to disk converison method")

                # try the last method, saving to disk and reading again

                opencv_image = qimage_tocv2_by_disk(qimage, logger, file_path="__img__.png")

                return opencv_image


class BaseDiagramWidget(QSplitter):
    """
    Common diagram widget to host common functions
    for the schematic and the map
    """

    def __init__(self,
                 gui: VeraGridMainGUI | DiagramsMain,
                 circuit: MultiCircuit,
                 diagram: Union[SchematicDiagram, MapDiagram],
                 library_model: Union[MapLibraryModel, SchematicLibraryModel],
                 time_index: Union[None, int] = None):
        """
        Constructor
        :param circuit:
        :param diagram:
        :param time_index:
        """
        QSplitter.__init__(self)

        self.gui = gui

        # --------------------------------------------------------------------------------------------------------------
        # Widget creation
        # --------------------------------------------------------------------------------------------------------------
        # Widget layout and child widgets:
        self.horizontal_layout = QHBoxLayout(self)

        # Table to display object's properties
        self.object_editor_table = QTableView(self)
        # change_font_size(self.object_editor_table, 9)
        # change_font_size(self.object_editor_table.verticalHeader(), 9)
        # change_font_size(self.object_editor_table.horizontalHeader(), 9)

        # Actual libraryView object
        self.library_view = QListView(self)
        self.library_view.setViewMode(self.library_view.ViewMode.ListMode)
        self.library_view.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        # change_font_size(self.library_view, 9)

        # library model
        self.library_model = library_model
        self.library_view.setModel(self.library_model)

        # create the grid name editor
        self.frame1 = QFrame()
        self.frame1_layout = QVBoxLayout()
        self.frame1_layout.setContentsMargins(0, 0, 0, 0)

        self.frame1_layout.addWidget(self.library_view)
        self.frame1.setLayout(self.frame1_layout)

        # Add the two objects into a layout
        splitter2 = QSplitter(self)
        splitter2.addWidget(self.frame1)
        splitter2.addWidget(self.object_editor_table)
        splitter2.setOrientation(Qt.Orientation.Vertical)
        self.addWidget(splitter2)
        # self.addWidget(self.editor_graphics_view)

        # factor 1:10
        splitter2.setStretchFactor(0, 2)
        splitter2.setStretchFactor(1, 5)

        # self.setStretchFactor(0, 0)
        # self.setStretchFactor(1, 2000)
        # --------------------------------------------------------------------------------------------------------------

        # store a reference to the multi circuit instance
        self.circuit: MultiCircuit = circuit

        # diagram to store the objects locations
        self.diagram: Union[SchematicDiagram, MapDiagram] = diagram

        # class to handle the relationships between widgets and API objects
        self.graphics_manager = GraphicsManager()

        # current time index from the GUI (None or 0, 1, 2, ..., n-1)
        self._time_index: Union[None, int] = time_index

        # logger
        self.logger: Logger = Logger()

        self.results_dictionary: Dict[SimulationTypes, DRIVER_OBJECTS] = dict()

        # video pointer
        self._video: Union[None, cv2.VideoWriter] = None

    def items(self) -> Generator[ALL_GRAPHICS, None, None]:
        """
        Iterable through all graphics registered in the graphics manager
        :return: ALL_GRAPHICS one by one
        """
        for device_tpe, graphics_dict in self.graphics_manager.graphic_dict.items():
            for idtag, graphical_obj in graphics_dict.items():
                yield graphical_obj

    @property
    def name(self):
        """
        Get the diagram name
        :return:
        """
        return self.diagram.name

    @name.setter
    def name(self, val: str):
        """
        Name setter
        :param val:
        :return:
        """
        self.diagram.name = val

    def _get_selected(self) -> List[GenericDiagramWidget]:
        """

        :return:
        """
        print(f"'get_selected' Not implemented for {str(self)}")
        return list()

    def _get_selection_api_objects(self) -> List[ALL_DEV_TYPES]:
        """
        Get a list of the API objects from the selection
        :return: List[ALL_DEV_TYPES]
        """
        return list()

    def _remove_from_scene(self, graphic_object: QGraphicsItem | GenericDiagramWidget) -> None:
        """
        Remove item from the diagram scene
        :param graphic_object: Graphic object associated
        """
        print(f"'remove_from_scene' Not implemented for {str(self)}")

    def remove_element(self,
                       device: ALL_DEV_TYPES,
                       graphic_object: GenericDiagramWidget | QGraphicsItem | None = None,
                       delete_from_db: bool = False) -> bool:
        """
        Remove device from the diagram and the database.
        If removing from the database, this propagates to all diagrams
        :param device: EditableDevice
        :param graphic_object: optionally provide the graphics object associated
        :param delete_from_db: Delete the element also from the database?
        :return: True if managed to delete_with_dialogue the object
        """
        if graphic_object is not None and device is not None:

            # Unregister this object from other objects that have references of it
            # i.e. unregister a line from the 2 buses that host connections to it
            # i.e. unregister a load from the bus that points to it
            graphic_object.delete_from_associations()

            if delete_from_db:
                self.circuit.delete_element(obj=device)

            # # For any other associated, graphic, delete too
            # for child_graphic in graphic_object.get_associated_widgets():
            #
            #     if delete_from_db:
            #         self.circuit.delete_element(obj=child_graphic.api_object)
            #
            #     # Warning: recursive call for devices that may have further sub-graphics (i.e. the nexus)
            #     self.remove_element(device=child_graphic.api_object,
            #                         graphic_object=child_graphic,
            #                         delete_from_db=delete_from_db)

            # Delete any other QWidget that is associated to this, and that we don't know about explicitly
            # i.e. the nexus of the loads, generators, etc...
            for child_graphic in graphic_object.get_extra_graphics():
                # simpler graphics associated, simply delete_with_dialogue
                self._remove_from_scene(graphic_object=child_graphic)

            # NOTE: This function already deleted from the database and other diagrams
            self.delete_element_utility_function(device=device, propagate=delete_from_db)
            self.object_editor_table.setModel(None)

            return True

        if graphic_object is None and device is not None:

            if delete_from_db:
                self.circuit.delete_element(obj=device)
                self.delete_element_utility_function(device=device, propagate=delete_from_db)
                self.object_editor_table.setModel(None)

            else:
                pass

            return True

        else:
            self.gui.show_warning_toast(f"Graphic object {graphic_object} and device {device} are none")
            self.object_editor_table.setModel(None)
            return False

    def delete_element_utility_function(self, device: ALL_DEV_TYPES, propagate: bool = True,
                                        graphic_object: ALL_GRAPHICS | None = None):
        """
        This function is a utility function to call this function in other diagrams through the GUI
        :param device: ALL_DEV_TYPES
        :param propagate: propagate
        :param graphic_object: QGraphicsItem
        :return:
        """
        self.diagram.delete_device(device=device)

        if graphic_object is None:
            graphic_object: QGraphicsItem = self.graphics_manager.delete_device(device=device)

        if graphic_object is not None:
            self._remove_from_scene(graphic_object)

            for extra_grph in graphic_object.get_extra_graphics():
                self._remove_from_scene(extra_grph)

        if propagate:
            self.gui.call_delete_db_element(caller=self, api_obj=device)

    def delete_with_dialogue(self, selected: List[GenericDiagramWidget], delete_from_db: bool) -> Tuple[bool, bool]:
        """
        Delete elements with a dialogue of all the dependencies
        :param selected: list of selected widgets
        :param delete_from_db: initial value for the delete from db option
        :return deleted? delete_from_db?
        """
        if len(selected) > 0:

            # get the set of all affected GenericDiagramWidget instances
            extended: Set[ALL_DEV_TYPES] = set()

            for graphic_obj in selected:

                if graphic_obj is not None:
                    if isinstance(graphic_obj, GenericDiagramWidget):
                        extended.add(graphic_obj.api_object)

                    for child_item in graphic_obj.get_associated_devices():
                        if child_item is not None:
                            extended.add(child_item)

            extended_lst: List[ALL_DEV_TYPES] = list(extended)

            dlg = DeleteDialogue(
                names_list=[f"{device.device_type.value}: "
                            f"{device.name}"
                            for device in extended_lst],
                delete_from_db=delete_from_db,
                title="Delete Selected",
                checks=False,
            )

            dlg.setModal(True)
            dlg.exec()

            if dlg.is_accepted:
                for device in extended_lst:
                    self.remove_element(device=device,
                                        graphic_object=self.graphics_manager.query(elm=device),
                                        delete_from_db=dlg.delete_from_db)

                return True, dlg.delete_from_db
            else:
                return False, False
        else:
            self.gui.show_warning_toast("Choose some elements to delete_with_dialogue")
            return False, False

    def delete_selected_from_widget(self, delete_from_db: bool) -> None:
        """
        Delete the selected items from the diagram
        :param delete_from_db:
        """
        self.delete_with_dialogue(selected=self._get_selected(),
                                  delete_from_db=delete_from_db)

    def delete_diagram_elements(self, elements: List[ALL_DEV_TYPES]):
        """
        Delete device from the diagram registry
        :param elements: list of elements to delete
        """
        for elm in elements:
            graphic_object: QGraphicsItem = self.graphics_manager.delete_device(device=elm)

            # this calls internally to delete_element_utility_function
            self.remove_element(
                device=elm,
                graphic_object=graphic_object
            )

            self.delete_element_utility_function(elm,
                                                 graphic_object=graphic_object)

    def set_time_index(self, time_index: Union[int, None]):
        """
        Set the time index of the table
        :param time_index: None or integer value
        """
        self._time_index = time_index

        mdl = self.object_editor_table.model()
        if isinstance(mdl, ObjectsModel):
            mdl.set_time_index(time_index=self._time_index)

    def get_time_index(self) -> Union[int, None]:
        """
        Get the time index
        :return: int, None
        """
        return self._time_index

    def set_editor_model(self, api_object: ALL_DEV_TYPES):
        """
        Set an api object to appear in the editable table view of the editor
        :param api_object: any EditableDevice
        """
        template_elm, dictionary_of_lists = self.circuit.get_dictionary_of_lists(api_object.device_type)
        mdl = ObjectsModel(objects=[api_object],
                           property_list=api_object.property_list,
                           time_index=self.get_time_index(),
                           parent=self.object_editor_table,
                           editable=True,
                           transposed=True,
                           dictionary_of_lists=dictionary_of_lists)

        self.object_editor_table.setModel(mdl)

    def set_results_to_plot(self, all_threads: List[DRIVER_OBJECTS]):
        """

        :param all_threads:
        :return:
        """
        self.results_dictionary = {thr.tpe: thr for thr in all_threads if thr is not None}

    def plot_branch(self, i: int, api_object: Union[Line, DcLine, Transformer2W, VSC, UPFC]):
        """
        Plot branch results
        :param i: branch index (not counting HVDC lines because those are not real Branches)
        :param api_object: API object
        """
        fig = plt.figure(figsize=(12, 8))
        fig.suptitle(api_object.name, fontsize=20)

        ax_1 = fig.add_subplot(211)
        ax_1.set_title('Probability x < value', fontsize=14)
        ax_1.set_ylabel('Loading [%]', fontsize=11)

        ax_2 = fig.add_subplot(212)
        ax_2.set_title('Power', fontsize=14)
        ax_2.set_ylabel('Power [MW]', fontsize=11)

        any_plot = False

        for driver, results in self.gui.session.drivers_results_iter():

            if results is not None:

                if isinstance(results, PowerFlowTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.BranchActivePowerFrom)
                    Sf_table.plot_device(ax=ax_1, device_idx=i, title="Power flow")

                    loading_table = results.mdl(result_type=ResultTypes.BranchLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot_device(ax=ax_2, device_idx=i, title="Power loading")
                    any_plot = True

                elif isinstance(results, LinearAnalysisTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.BranchActivePowerFrom)
                    Sf_table.plot_device(ax=ax_1, device_idx=i, title="Linear flow")

                    loading_table = results.mdl(result_type=ResultTypes.BranchLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot_device(ax=ax_2, device_idx=i, title="Linear loading")
                    any_plot = True

                elif isinstance(results, ContingencyAnalysisTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.MaxContingencyFlows)
                    Sf_table.plot_device(ax=ax_1, device_idx=i, title="Contingency flow")

                    loading_table = results.mdl(result_type=ResultTypes.MaxContingencyLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot_device(ax=ax_2, device_idx=i, title="Contingency loading")
                    any_plot = True

                elif isinstance(results, OptimalPowerFlowTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.BranchActivePowerFrom)
                    Sf_table.plot_device(ax=ax_1, device_idx=i, title="Optimal power flow")

                    loading_table = results.mdl(result_type=ResultTypes.BranchLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot_device(ax=ax_2, device_idx=i, title="Optimal loading")
                    any_plot = True

                elif isinstance(results, StochasticPowerFlowResults):
                    loading_table = results.mdl(result_type=ResultTypes.BranchLoadingCDF)
                    loading_table.convert_to_cdf()
                    loading_table.plot_device(ax=ax_2, device_idx=i, title="Stochastic loading")
                    any_plot = True

        if any_plot:
            plt.legend()
            plt.show()
        else:
            info_msg("No time series results to plot, run some time series results. Even partial results are fine",
                     f"{api_object.name} results plot")

    def plot_hvdc_branch(self, i: int, api_object: HvdcLine):
        """
        HVDC branch
        :param i: index of the object
        :param api_object: HvdcGraphicItem
        """
        fig = plt.figure(figsize=(12, 8))
        fig.suptitle(api_object.name, fontsize=20)

        ax_1 = fig.add_subplot(211)
        ax_1.set_title('Probability x < value', fontsize=14)
        ax_1.set_ylabel('Loading [%]', fontsize=11)

        ax_2 = fig.add_subplot(212)
        ax_2.set_title('Power', fontsize=14)
        ax_2.set_ylabel('Power [MW]', fontsize=11)

        any_plot = False

        for driver, results in self.gui.session.drivers_results_iter():

            if results is not None:

                if isinstance(results, PowerFlowTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.HvdcPowerFrom)
                    Sf_table.plot(ax=ax_1, selected_col_idx=[i])

                    loading_table = results.mdl(result_type=ResultTypes.HvdcLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot(ax=ax_2, selected_col_idx=[i])
                    any_plot = True

                elif isinstance(results, OptimalPowerFlowTimeSeriesResults):

                    Sf_table = results.mdl(result_type=ResultTypes.HvdcPowerFrom)
                    Sf_table.plot(ax=ax_1, selected_col_idx=[i])

                    loading_table = results.mdl(result_type=ResultTypes.HvdcLoading)
                    loading_table.convert_to_cdf()
                    loading_table.plot(ax=ax_2, selected_col_idx=[i])
                    any_plot = True

        if any_plot:
            plt.legend()
            plt.show()
        else:
            info_msg("No time series results to plot, run some time series results. Even partial results are fine",
                     f"{api_object.name} results plot")

    @staticmethod
    def set_rate_to_profile(api_object: ALL_DEV_TYPES):
        """

        :param api_object:
        """
        if api_object is not None:
            if api_object.rate_prof.size():
                quit_msg = (f"{api_object.name}\nAre you sure that you want to overwrite the "
                            f"rates profile with the snapshot value?")

                ok = yes_no_question(text=quit_msg, title='Overwrite the profile')

                if ok:
                    api_object.rate_prof.fill(api_object.rate)

    @staticmethod
    def set_active_status_to_profile(api_object: ALL_DEV_TYPES, override_question=False):
        """

        :param api_object:
        :param override_question:
        :return:
        """
        if api_object is not None:
            if api_object.active_prof.size():
                if not override_question:
                    quit_msg = (f"{api_object.name}\nAre you sure that you want to overwrite the "
                                f"active profile with the snapshot value?")

                    ok = yes_no_question(text=quit_msg, title='Overwrite the active profile')
                else:
                    ok = True

                if ok:
                    if api_object.active:
                        api_object.active_prof.fill(True)
                    else:
                        api_object.active_prof.fill(False)

    def draw(self) -> None:
        """
        Draw the stored diagram
        """
        self.draw_diagram(diagram=self.diagram)

    def draw_diagram(self, diagram: Union[SchematicDiagram, MapDiagram]) -> None:
        """
        Draw the diagram
        :param diagram: Map or schematic diagram
        """
        pass

    def clear(self) -> None:
        """
        Clear the schematic
        """
        self.graphics_manager.clear()

    def set_data(self, circuit: MultiCircuit, diagram: SchematicDiagram):
        """
        Set the widget data and redraw
        :param circuit: MultiCircuit
        :param diagram: SchematicDiagram
        """
        self.clear()
        self.circuit = circuit
        self.diagram = diagram
        self.draw()

    def colour_results(self,
                       Sbus: CxVec,
                       bus_active: IntVec,
                       Sf: CxVec,
                       St: CxVec,
                       voltages: CxVec,
                       loadings: CxVec,
                       types: IntVec = None,
                       losses: CxVec = None,
                       br_active: IntVec = None,
                       hvdc_Pf: Vec = None,
                       hvdc_Pt: Vec = None,
                       hvdc_losses: Vec = None,
                       hvdc_loading: Vec = None,
                       hvdc_active: IntVec = None,
                       loading_label: str = 'loading',
                       vsc_Pf: Vec = None,
                       vsc_Pt: Vec = None,
                       vsc_Qt: Vec = None,
                       vsc_losses: Vec = None,
                       vsc_loading: Vec = None,
                       vsc_active: IntVec = None,
                       ma: Vec = None,
                       tau: Vec = None,
                       fluid_node_p2x_flow: Vec = None,
                       fluid_node_current_level: Vec = None,
                       fluid_node_spillage: Vec = None,
                       fluid_node_flow_in: Vec = None,
                       fluid_node_flow_out: Vec = None,
                       fluid_path_flow: Vec = None,
                       fluid_injection_flow: Vec = None,
                       use_flow_based_width: bool = False,
                       min_branch_width: int = 5,
                       max_branch_width=5,
                       min_bus_width=20,
                       max_bus_width=20,
                       cmap: palettes.Colormaps = None,
                       is_three_phase: bool = False):
        """

        :param Sbus:
        :param bus_active:
        :param Sf:
        :param St:
        :param voltages:
        :param loadings:
        :param types:
        :param losses:
        :param br_active:
        :param hvdc_Pf:
        :param hvdc_Pt:
        :param hvdc_losses:
        :param hvdc_loading:
        :param hvdc_active:
        :param loading_label:
        :param vsc_Pf:
        :param vsc_Pt:
        :param vsc_Qt:
        :param vsc_losses:
        :param vsc_loading:
        :param vsc_active:
        :param ma:
        :param tau:
        :param fluid_node_p2x_flow:
        :param fluid_node_current_level:
        :param fluid_node_spillage:
        :param fluid_node_flow_in:
        :param fluid_node_flow_out:
        :param fluid_path_flow:
        :param fluid_injection_flow:
        :param use_flow_based_width:
        :param min_branch_width:
        :param max_branch_width:
        :param min_bus_width:
        :param max_bus_width:
        :param cmap:
        :param is_three_phase: the results are three-phase
        :return:
        """
        pass

    def colour_results_3ph(self,
                           SbusA: CxVec,
                           SbusB: CxVec,
                           SbusC: CxVec,
                           voltagesA: CxVec,
                           voltagesB: CxVec,
                           voltagesC: CxVec,
                           bus_active: IntVec,
                           types: IntVec,
                           SfA: CxVec,
                           SfB: CxVec,
                           SfC: CxVec,
                           StA: CxVec,
                           StB: CxVec,
                           StC: CxVec,
                           loadingsA: CxVec,
                           loadingsB: CxVec,
                           loadingsC: CxVec,
                           lossesA: CxVec,
                           lossesB: CxVec,
                           lossesC: CxVec,
                           br_active: IntVec,
                           ma: Vec,
                           tau: Vec,
                           hvdc_PfA: Vec,
                           hvdc_PfB: Vec,
                           hvdc_PfC: Vec,
                           hvdc_PtA: Vec,
                           hvdc_PtB: Vec,
                           hvdc_PtC: Vec,
                           hvdc_losses: Vec,
                           hvdc_loading: Vec,
                           hvdc_active: IntVec,
                           vsc_Pf: Vec,
                           vsc_PtA: Vec,
                           vsc_PtB: Vec,
                           vsc_PtC: Vec,
                           vsc_QtA: Vec,
                           vsc_QtB: Vec,
                           vsc_QtC: Vec,
                           vsc_losses: Vec,
                           vsc_loading: Vec,
                           vsc_active: IntVec,
                           loading_label: str = 'loading',
                           use_flow_based_width: bool = False,
                           min_branch_width: int = 5,
                           max_branch_width=5,
                           min_bus_width=20,
                           max_bus_width=20,
                           cmap: palettes.Colormaps = None):
        pass

    def disable_all_results_tags(self):
        """
        Disable all results' tags in this diagram
        """
        for device_tpe, type_dict in self.graphics_manager.graphic_dict.items():
            for key, widget in type_dict.items():
                widget.disable_label_drawing()

    def enable_all_results_tags(self):
        """
        Enable all results' tags in this diagram
        """
        for device_tpe, type_dict in self.graphics_manager.graphic_dict.items():
            for key, widget in type_dict.items():
                widget.enable_label_drawing()

    def get_picture_width(self) -> int:
        """
        Width
        :return: width in pixels
        """
        return 0

    def get_picture_height(self) -> int:
        """
        Height
        :return: height in pixels
        """
        return 0

    def get_image(self, transparent: bool = False) -> QImage:
        """
        get the current picture
        :param transparent: Set a transparent background
        :return: QImage, width, height
        """
        pass

    def take_picture(self, filename: str):
        """
        Save the grid to a png file
        :param filename: Picture file name
        """
        pass

    def start_video_recording(self, fname: str, fps: int = 30, logger: Logger = Logger()) -> Tuple[int, int]:
        """
        Save video
        :param fname: file name
        :param fps: frames per second
        :param logger: LOgger
        :returns width, height
        """

        image = self.get_image()
        w = image.width()
        h = image.height()
        cv2_image = qimage_to_cv(image, logger)
        w2, h2, _ = cv2_image.shape

        if fname.endswith('.mp4'):
            self._video = cv2.VideoWriter(filename=fname,
                                          fourcc=cv2.VideoWriter_fourcc(*'mp4v'),
                                          fps=fps,
                                          frameSize=(w, h))
        elif fname.endswith('.avi'):
            self._video = cv2.VideoWriter(filename=fname + '.avi',
                                          fourcc=cv2.VideoWriter_fourcc(*'XVID'),
                                          fps=fps,
                                          frameSize=(w, h))
        else:
            raise Exception(f"File format not recognized {fname}")

        return w, h

    def capture_video_frame(self, w: int, h: int, logger: Logger):
        """
        Save video frame
        """
        image = self.get_image()
        w2 = image.width()
        h2 = image.height()

        if w != w2:
            logger.add_error(f"Width {w2} different from expected width {w}")

        if h != h2:
            logger.add_error(f"Height {h2} different from expected width {h}")

        cv2_image = qimage_to_cv(image, logger)

        if cv2_image is not None:
            self._video.write(cv2_image)

    def end_video_recording(self):
        """
        Finalize video recording
        """
        self._video.release()
        print("Video released")

    def update_label_drwaing_status(self, device: ALL_DEV_TYPES, draw_labels: bool) -> None:
        """
        Update the label drawing flag
        :param device: Any database device
        :param draw_labels: Draw labels?
        """
        location = self.diagram.query_point(device=device)

        if location is not None:
            location.draw_labels = draw_labels

    def set_size_constraints(self,
                             use_flow_based_width: bool = False,
                             min_branch_width: int = 5,
                             max_branch_width=5,
                             min_bus_width=20,
                             max_bus_width=20,
                             arrow_size=20):
        """
        Set the size constraints
        :param use_flow_based_width:
        :param min_branch_width:
        :param max_branch_width:
        :param min_bus_width:
        :param max_bus_width:
        :param arrow_size:
        """
        self.diagram.set_size_constraints(
            use_flow_based_width=use_flow_based_width,
            min_branch_width=min_branch_width,
            max_branch_width=max_branch_width,
            min_bus_width=min_bus_width,
            max_bus_width=max_bus_width,
            arrow_size=arrow_size
        )

    def copy(self):
        """

        :return:
        """
        raise Exception('Copy method not implemented!')

    def consolidate_coordinates(self):
        """

        :return:
        """
        raise Exception('Consolidate method method not implemented!')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import re
from PySide6.QtGui import (
    QSyntaxHighlighter,
    QTextCharFormat,
    QColor,
    QFont,
)


class PythonHighlighter(QSyntaxHighlighter):
    """
    High-quality Python syntax highlighter for interactive consoles.
    """

    NORMAL = 0
    TRIPLE_SINGLE = 1
    TRIPLE_DOUBLE = 2

    def __init__(self, document):
        super().__init__(document)

        # ----------------------------
        # Formats (created ONCE)
        # ----------------------------

        def fmt(color, bold=False):
            f = QTextCharFormat()
            f.setForeground(QColor(color))
            if bold:
                f.setFontWeight(QFont.Weight.Bold)
            return f

        self.formats = {
            "prompt": fmt("#64B5F6", True),
            "keyword": fmt("#82AAFF", True),
            "builtin": fmt("#FFCB6B"),
            "number": fmt("#C3E88D"),
            "string": fmt("#F07178"),
            "comment": fmt("#546E7A"),
            "operator": fmt("#89DDFF"),
            "identifier": fmt("#FFFFFF"),
            "defclass": fmt("#C792EA", True),
        }

        # ----------------------------
        # Keyword sets
        # ----------------------------

        self.keywords = {
            "and", "as", "assert", "break", "class", "continue",
            "def", "del", "elif", "else", "except", "False",
            "finally", "for", "from", "global", "if", "import",
            "in", "is", "lambda", "None", "nonlocal", "not",
            "or", "pass", "raise", "return", "True", "try",
            "while", "with", "yield",
        }

        self.builtins = {
            "print", "len", "range", "open", "min", "max",
            "sum", "abs", "round", "zip", "enumerate", "sorted",
            "map", "filter", "list", "dict", "set", "tuple",
            "int", "float", "str", "bool",
        }

        # ----------------------------
        # Precompiled regexes
        # ----------------------------

        self.re_prompt = re.compile(r"^(>>> |\.\.\. )")
        self.re_identifier = re.compile(r"\b[A-Za-z_]\w*\b")
        self.re_number = re.compile(r"\b\d+(\.\d+)?([eE][+-]?\d+)?\b")
        self.re_comment = re.compile(r"#.*$")
        self.re_operator = re.compile(r"[+\-*/%=<>!&|^~]+")

        self.re_string_single = re.compile(r"'([^'\\]|\\.)*'")
        self.re_string_double = re.compile(r'"([^"\\]|\\.)*"')

        self.re_triple_single = re.compile(r"'''")
        self.re_triple_double = re.compile(r'"""')

    # ----------------------------
    # Core highlighter
    # ----------------------------

    def highlightBlock(self, text):
        self.setCurrentBlockState(self.NORMAL)

        # ---- Prompt ----
        m = self.re_prompt.match(text)
        if m:
            self.setFormat(0, m.end(), self.formats["prompt"])

        # ---- Multiline strings ----
        if self._handle_multiline_strings(text):
            return

        # ---- Comments ----
        for m in self.re_comment.finditer(text):
            self.setFormat(m.start(), m.end() - m.start(), self.formats["comment"])

        # ---- Strings ----
        for rx in (self.re_string_single, self.re_string_double):
            for m in rx.finditer(text):
                self.setFormat(m.start(), m.end() - m.start(), self.formats["string"])

        # ---- Numbers ----
        for m in self.re_number.finditer(text):
            self.setFormat(m.start(), m.end() - m.start(), self.formats["number"])

        # ---- Operators ----
        for m in self.re_operator.finditer(text):
            self.setFormat(m.start(), m.end() - m.start(), self.formats["operator"])

        # ---- Identifiers, keywords, builtins ----
        for m in self.re_identifier.finditer(text):
            word = m.group()
            if word in self.keywords:
                self.setFormat(m.start(), len(word), self.formats["keyword"])
            elif word in self.builtins:
                self.setFormat(m.start(), len(word), self.formats["builtin"])

    # ----------------------------
    # Multiline string handling
    # ----------------------------

    def _handle_multiline_strings(self, text):
        if self.previousBlockState() == self.TRIPLE_SINGLE:
            end = self.re_triple_single.search(text)
            if end:
                self.setFormat(0, end.end(), self.formats["string"])
                self.setCurrentBlockState(self.NORMAL)
                return False
            else:
                self.setFormat(0, len(text), self.formats["string"])
                self.setCurrentBlockState(self.TRIPLE_SINGLE)
                return True

        if self.previousBlockState() == self.TRIPLE_DOUBLE:
            end = self.re_triple_double.search(text)
            if end:
                self.setFormat(0, end.end(), self.formats["string"])
                self.setCurrentBlockState(self.NORMAL)
                return False
            else:
                self.setFormat(0, len(text), self.formats["string"])
                self.setCurrentBlockState(self.TRIPLE_DOUBLE)
                return True

        start = self.re_triple_single.search(text)
        if start:
            self.setFormat(start.start(), len(text) - start.start(), self.formats["string"])
            self.setCurrentBlockState(self.TRIPLE_SINGLE)
            return True

        start = self.re_triple_double.search(text)
        if start:
            self.setFormat(start.start(), len(text) - start.start(), self.formats["string"])
            self.setCurrentBlockState(self.TRIPLE_DOUBLE)
            return True

        return False
